# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Automatic concatenation of multiple cubes over one or more existing dimensions.

"""

from collections import defaultdict, namedtuple

import dask.array as da
import numpy as np

import iris.coords
import iris.cube
from iris.util import guess_coord_axis, array_equal


#
# TODO:
#
#   * Cope with auxiliary coordinate factories.
#
#   * Allow concatentation over a user specified dimension.
#


# Restrict the names imported from this namespace.
__all__ = ["concatenate"]

# Direction of dimension coordinate value order.
_CONSTANT = 0
_DECREASING = -1
_INCREASING = 1


class _CoordAndDims(namedtuple("CoordAndDims", ["coord", "dims"])):
    """
    Container for a coordinate and the associated data dimension(s)
    spanned over a :class:`iris.cube.Cube`.

    Args:

    * coord:
        A :class:`iris.coords.DimCoord` or :class:`iris.coords.AuxCoord`
        coordinate instance.

    * dims:
        A tuple of the data dimension(s) spanned by the coordinate.

    """

    __slots__ = ()


class _CoordMetaData(
    namedtuple(
        "CoordMetaData",
        ["defn", "dims", "points_dtype", "bounds_dtype", "kwargs"],
    )
):
    """
    Container for the metadata that defines a dimension or auxiliary
    coordinate.

    Args:

    * defn:
        The :class:`iris.common.CoordMetadata` metadata that represents a
        coordinate.

    * dims:
        The dimension(s) associated with the coordinate.

    * points_dtype:
        The points data :class:`np.dtype` of an associated coordinate.

    * bounds_dtype:
        The bounds data :class:`np.dtype` of an associated coordinate.

    * kwargs:
        A dictionary of key/value pairs required to define a coordinate.

    """

    def __new__(mcs, coord, dims):
        """
        Create a new :class:`_CoordMetaData` instance.

        Args:

        * coord:
            The :class:`iris.coord.DimCoord` or :class:`iris.coord.AuxCoord`.

        * dims:
            The dimension(s) associated with the coordinate.

        Returns:
            The new class instance.

        """
        defn = coord.metadata
        points_dtype = coord.points.dtype
        bounds_dtype = coord.bounds.dtype if coord.bounds is not None else None
        kwargs = {}
        # Add scalar flag metadata.
        kwargs["scalar"] = coord.points.size == 1
        # Add circular flag metadata for dimensional coordinates.
        if hasattr(coord, "circular"):
            kwargs["circular"] = coord.circular
        if isinstance(coord, iris.coords.DimCoord):
            # Mix the monotonic ordering into the metadata.
            if coord.points[0] == coord.points[-1]:
                order = _CONSTANT
            elif coord.points[-1] > coord.points[0]:
                order = _INCREASING
            else:
                order = _DECREASING
            kwargs["order"] = order
        metadata = super().__new__(
            mcs, defn, dims, points_dtype, bounds_dtype, kwargs
        )
        return metadata

    __slots__ = ()

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        result = NotImplemented
        if isinstance(other, _CoordMetaData):
            sprops, oprops = self._asdict(), other._asdict()
            # Ignore "kwargs" meta-data for the first comparison.
            sprops["kwargs"] = oprops["kwargs"] = None
            result = sprops == oprops
            if result:
                skwargs, okwargs = self.kwargs.copy(), other.kwargs.copy()
                # Monotonic "order" only applies to DimCoord's.
                # The monotonic "order" must be _INCREASING or _DECREASING if
                # the DimCoord is NOT "scalar". Otherwise, if the DimCoord is
                # "scalar" then the "order" must be _CONSTANT.
                if skwargs["scalar"] or okwargs["scalar"]:
                    # We don't care about the monotonic "order" given that
                    # at least one coordinate is a scalar coordinate.
                    skwargs["scalar"] = okwargs["scalar"] = None
                    skwargs["order"] = okwargs["order"] = None
                result = skwargs == okwargs
        return result

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is not NotImplemented:
            result = not result
        return result

    def name(self):
        """Get the name from the coordinate definition."""
        return self.defn.name()


class _OtherMetaData(namedtuple("OtherMetaData", ["defn", "dims"])):
    """
    Container for the metadata that defines a cell measure or ancillary
    variable.

    Args:

    * defn:
        The :class:`iris.coords._DMDefn` or :class:`iris.coords._CellMeasureDefn`
        metadata that represents a coordinate.

    * dims:
        The dimension(s) associated with the coordinate.

    """

    def __new__(cls, ancil, dims):
        """
        Create a new :class:`_OtherMetaData` instance.

        Args:

        * ancil:
            The :class:`iris.coord.CellMeasure` or
            :class:`iris.coord.AncillaryVariable`.

        * dims:
            The dimension(s) associated with ancil.

        Returns:
            The new class instance.

        """
        defn = ancil.metadata
        metadata = super().__new__(cls, defn, dims)
        return metadata

    __slots__ = ()

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        result = NotImplemented
        if isinstance(other, _OtherMetaData):
            result = self._asdict() == other._asdict()
        return result

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is not NotImplemented:
            result = not result
        return result

    def name(self):
        """Get the name from the coordinate definition."""
        return self.defn.name()


class _SkeletonCube(namedtuple("SkeletonCube", ["signature", "data"])):
    """
    Basis of a source-cube, containing the associated coordinate metadata,
    coordinates and cube data payload.

    Args:

    * signature:
        The :class:`_CoordSignature` of an associated source-cube.

    * data:
        The data payload of an associated :class:`iris.cube.Cube` source-cube.

    """

    __slots__ = ()


class _Extent(namedtuple("Extent", ["min", "max"])):
    """
    Container representing the limits of a one-dimensional extent/range.

    Args:

    * min:
        The minimum value of the extent.

    * max:
        The maximum value of the extent.

    """

    __slots__ = ()


class _CoordExtent(namedtuple("CoordExtent", ["points", "bounds"])):
    """
    Container representing the points and bounds extent of a one dimensional
    coordinate.

    Args:

    * points:
        The :class:`_Extent` of the coordinate point values.

    * bounds:
        A list containing the :class:`_Extent` of the coordinate lower
        bound and the upper bound. Defaults to None if no associated
        bounds exist for the coordinate.

    """

    __slots__ = ()


def concatenate(
    cubes,
    error_on_mismatch=False,
    check_aux_coords=True,
    check_cell_measures=True,
    check_ancils=True,
):
    """
    Concatenate the provided cubes over common existing dimensions.

    Args:

    * cubes:
        An iterable containing one or more :class:`iris.cube.Cube` instances
        to be concatenated together.

    Kwargs:

    * error_on_mismatch:
        If True, raise an informative
        :class:`~iris.exceptions.ContatenateError` if registration fails.

    Returns:
        A :class:`iris.cube.CubeList` of concatenated :class:`iris.cube.Cube`
        instances.

    """
    proto_cubes_by_name = defaultdict(list)
    # Initialise the nominated axis (dimension) of concatenation
    # which requires to be negotiated.
    axis = None

    # Register each cube with its appropriate proto-cube.
    for cube in cubes:
        name = cube.standard_name or cube.long_name
        proto_cubes = proto_cubes_by_name[name]
        registered = False

        # Register cube with an existing proto-cube.
        for proto_cube in proto_cubes:
            registered = proto_cube.register(
                cube,
                axis,
                error_on_mismatch,
                check_aux_coords,
                check_cell_measures,
                check_ancils,
            )
            if registered:
                axis = proto_cube.axis
                break

        # Create a new proto-cube for an unregistered cube.
        if not registered:
            proto_cubes.append(_ProtoCube(cube))

    # Construct a concatenated cube from each of the proto-cubes.
    concatenated_cubes = iris.cube.CubeList()

    # Emulate Python 2 behaviour.
    def _none_sort(item):
        return (item is not None, item)

    for name in sorted(proto_cubes_by_name, key=_none_sort):
        for proto_cube in proto_cubes_by_name[name]:
            # Construct the concatenated cube.
            concatenated_cubes.append(proto_cube.concatenate())

    # Perform concatenation until we've reached an equilibrium.
    count = len(concatenated_cubes)
    if count != 1 and count != len(cubes):
        concatenated_cubes = concatenate(concatenated_cubes)

    return concatenated_cubes


class _CubeSignature:
    """
    Template for identifying a specific type of :class:`iris.cube.Cube` based
    on its metadata, coordinates and cell_measures.

    """

    def __init__(self, cube):
        """
        Represents the cube metadata and associated coordinate metadata that
        allows suitable cubes for concatenation to be identified.

        Args:

        * cube:
            The :class:`iris.cube.Cube` source-cube.

        """
        self.aux_coords_and_dims = []
        self.aux_metadata = []
        self.dim_coords = cube.dim_coords
        self.dim_metadata = []
        self.ndim = cube.ndim
        self.scalar_coords = []
        self.cell_measures_and_dims = []
        self.cm_metadata = []
        self.ancillary_variables_and_dims = []
        self.av_metadata = []
        self.dim_mapping = []

        # Determine whether there are any anonymous cube dimensions.
        covered = set(cube.coord_dims(coord)[0] for coord in self.dim_coords)
        self.anonymous = covered != set(range(self.ndim))

        self.defn = cube.metadata
        self.data_type = cube.dtype

        #
        # Collate the dimension coordinate metadata.
        #
        for ind, coord in enumerate(self.dim_coords):
            dims = cube.coord_dims(coord)
            metadata = _CoordMetaData(coord, dims)
            self.dim_metadata.append(metadata)
            self.dim_mapping.append(dims[0])

        #
        # Collate the auxiliary coordinate metadata and scalar coordinates.
        #
        axes = dict(T=0, Z=1, Y=2, X=3)

        # Coordinate sort function - by guessed coordinate axis, then
        # by coordinate name, then by dimensions, in ascending order.
        def key_func(coord):
            return (
                axes.get(guess_coord_axis(coord), len(axes) + 1),
                coord.name(),
                cube.coord_dims(coord),
            )

        for coord in sorted(cube.aux_coords, key=key_func):
            dims = cube.coord_dims(coord)
            if dims:
                metadata = _CoordMetaData(coord, dims)
                self.aux_metadata.append(metadata)
                coord_and_dims = _CoordAndDims(coord, tuple(dims))
                self.aux_coords_and_dims.append(coord_and_dims)
            else:
                self.scalar_coords.append(coord)

        def meta_key_func(dm):
            return (dm.metadata, dm.cube_dims(cube))

        for cm in sorted(cube.cell_measures(), key=meta_key_func):
            dims = cube.cell_measure_dims(cm)
            metadata = _OtherMetaData(cm, dims)
            self.cm_metadata.append(metadata)
            cm_and_dims = _CoordAndDims(cm, tuple(dims))
            self.cell_measures_and_dims.append(cm_and_dims)

        for av in sorted(cube.ancillary_variables(), key=meta_key_func):
            dims = cube.ancillary_variable_dims(av)
            metadata = _OtherMetaData(av, dims)
            self.av_metadata.append(metadata)
            av_and_dims = _CoordAndDims(av, tuple(dims))
            self.ancillary_variables_and_dims.append(av_and_dims)

    def _coordinate_differences(self, other, attr):
        """
        Determine the names of the coordinates that differ between `self` and
        `other` for a coordinate attribute on a _CubeSignature.

        Args:

        * other (_CubeSignature):
            The _CubeSignature to compare against.

        * attr (string):
            The _CubeSignature attribute within which differences exist
            between `self` and `other`.

        Returns:
            Tuple of a descriptive error message and the names of coordinates
            that differ between `self` and `other`.

        """
        # Set up {name: coord_metadata} dictionaries.
        self_dict = {x.name(): x for x in getattr(self, attr)}
        other_dict = {x.name(): x for x in getattr(other, attr)}
        if len(self_dict) == 0:
            self_dict = {"< None >": None}
        if len(other_dict) == 0:
            other_dict = {"< None >": None}
        self_names = sorted(self_dict.keys())
        other_names = sorted(other_dict.keys())

        # Compare coord metadata.
        if len(self_names) != len(other_names) or self_names != other_names:
            result = ("", ", ".join(self_names), ", ".join(other_names))
        else:
            diff_names = []
            for self_key, self_value in self_dict.items():
                other_value = other_dict[self_key]
                if self_value != other_value:
                    diff_names.append(self_key)
            result = (
                " metadata",
                ", ".join(diff_names),
                ", ".join(diff_names),
            )
        return result

    def match(self, other, error_on_mismatch):
        """
        Return whether this _CubeSignature equals another.

        This is the first step to determine if two "cubes" (either a
        real Cube or a ProtoCube) can be concatenated, by considering:
            - data dimensions
            - dimensions metadata
            - aux coords metadata
            - scalar coords
            - attributes
            - dtype

        Args:

        * other (_CubeSignature):
            The _CubeSignature to compare against.

        * error_on_mismatch (bool):
            If True, raise a :class:`~iris.exceptions.MergeException`
            with a detailed explanation if the two do not match.

        Returns:
           Boolean. True if and only if this _CubeSignature matches the other.

        """
        msg_template = "{}{} differ: {} != {}"
        msgs = []

        # Check cube definitions.
        if self.defn != other.defn:
            # Note that the case of different phenomenon names is dealt
            # with in :meth:`iris.cube.CubeList.concatenate_cube()`.
            msg = "Cube metadata differs for phenomenon: {}"
            msgs.append(msg.format(self.defn.name()))

        # Check dim coordinates.
        if self.dim_metadata != other.dim_metadata:
            differences = self._coordinate_differences(other, "dim_metadata")
            msgs.append(
                msg_template.format("Dimension coordinates", *differences)
            )
        # Check aux coordinates.
        if self.aux_metadata != other.aux_metadata:
            differences = self._coordinate_differences(other, "aux_metadata")
            msgs.append(
                msg_template.format("Auxiliary coordinates", *differences)
            )
        # Check cell measures.
        if self.cm_metadata != other.cm_metadata:
            differences = self._coordinate_differences(other, "cm_metadata")
            msgs.append(msg_template.format("Cell measures", *differences))
        # Check ancillary variables.
        if self.av_metadata != other.av_metadata:
            differences = self._coordinate_differences(other, "av_metadata")
            msgs.append(
                msg_template.format("Ancillary variables", *differences)
            )
        # Check scalar coordinates.
        if self.scalar_coords != other.scalar_coords:
            differences = self._coordinate_differences(other, "scalar_coords")
            msgs.append(
                msg_template.format("Scalar coordinates", *differences)
            )
        # Check ndim.
        if self.ndim != other.ndim:
            msgs.append(
                msg_template.format(
                    "Data dimensions", "", self.ndim, other.ndim
                )
            )
        # Check data type.
        if self.data_type != other.data_type:
            msgs.append(
                msg_template.format(
                    "Data types", "", self.data_type, other.data_type
                )
            )

        match = not bool(msgs)
        if error_on_mismatch and not match:
            raise iris.exceptions.ConcatenateError(msgs)
        return match


class _CoordSignature:
    """
    Template for identifying a specific type of :class:`iris.cube.Cube` based
    on its coordinates.

    """

    def __init__(self, cube_signature):
        """
        Represents the coordinate metadata required to identify suitable
        non-overlapping :class:`iris.cube.Cube` source-cubes for
        concatenation over a common single dimension.

        Args:

        * cube_signature:
            The :class:`_CubeSignature` that defines the source-cube.

        """
        self.aux_coords_and_dims = cube_signature.aux_coords_and_dims
        self.cell_measures_and_dims = cube_signature.cell_measures_and_dims
        self.ancillary_variables_and_dims = (
            cube_signature.ancillary_variables_and_dims
        )
        self.dim_coords = cube_signature.dim_coords
        self.dim_mapping = cube_signature.dim_mapping
        self.dim_extents = []
        self.dim_order = [
            metadata.kwargs["order"]
            for metadata in cube_signature.dim_metadata
        ]

        # Calculate the extents for each dimensional coordinate.
        self._calculate_extents()

    @staticmethod
    def _cmp(coord, other):
        """
        Compare the coordinates for concatenation compatibility.

        Returns:
            A boolean tuple pair of whether the coordinates are compatible,
            and whether they represent a candidate axis of concatenation.

        """
        # A candidate axis must have non-identical coordinate points.
        candidate_axis = not array_equal(coord.points, other.points)

        if candidate_axis:
            # Ensure both have equal availability of bounds.
            result = (coord.bounds is None) == (other.bounds is None)
        else:
            if coord.bounds is not None and other.bounds is not None:
                # Ensure equality of bounds.
                result = array_equal(coord.bounds, other.bounds)
            else:
                # Ensure both have equal availability of bounds.
                result = coord.bounds is None and other.bounds is None

        return result, candidate_axis

    def candidate_axis(self, other):
        """
        Determine the candidate axis of concatenation with the
        given coordinate signature.

        If a candidate axis is found, then the coordinate
        signatures are compatible.

        Args:

        * other:
            The :class:`_CoordSignature`

        Returns:
            None if no single candidate axis exists, otherwise
            the candidate axis of concatenation.

        """
        result = False
        candidate_axes = []

        # Compare dimension coordinates.
        for ind, coord in enumerate(self.dim_coords):
            result, candidate_axis = self._cmp(coord, other.dim_coords[ind])
            if not result:
                break
            if candidate_axis:
                dim = self.dim_mapping[ind]
                candidate_axes.append(dim)

        # Only permit one degree of dimensional freedom when
        # determining the candidate axis of concatenation.
        if result and len(candidate_axes) == 1:
            result = candidate_axes[0]
        else:
            result = None

        return result

    def _calculate_extents(self):
        """
        Calculate the extent over each dimension coordinates points and bounds.

        """
        self.dim_extents = []
        for coord, order in zip(self.dim_coords, self.dim_order):
            if order == _CONSTANT or order == _INCREASING:
                points = _Extent(coord.points[0], coord.points[-1])
                if coord.bounds is not None:
                    bounds = (
                        _Extent(coord.bounds[0, 0], coord.bounds[-1, 0]),
                        _Extent(coord.bounds[0, 1], coord.bounds[-1, 1]),
                    )
                else:
                    bounds = None
            else:
                # The order must be decreasing ...
                points = _Extent(coord.points[-1], coord.points[0])
                if coord.bounds is not None:
                    bounds = (
                        _Extent(coord.bounds[-1, 0], coord.bounds[0, 0]),
                        _Extent(coord.bounds[-1, 1], coord.bounds[0, 1]),
                    )
                else:
                    bounds = None

            self.dim_extents.append(_CoordExtent(points, bounds))


class _ProtoCube:
    """
    Framework for concatenating multiple source-cubes over one
    common dimension.

    """

    def __init__(self, cube):
        """
        Create a new _ProtoCube from the given cube and record the cube
        as a source-cube.

        Args:

        * cube:
            Source :class:`iris.cube.Cube` of the :class:`_ProtoCube`.

        """
        # Cache the source-cube of this proto-cube.
        self._cube = cube

        # The cube signature is a combination of cube and coordinate
        # metadata that defines this proto-cube.
        self._cube_signature = _CubeSignature(cube)

        # The coordinate signature allows suitable non-overlapping
        # source-cubes to be identified.
        self._coord_signature = _CoordSignature(self._cube_signature)

        # The list of source-cubes relevant to this proto-cube.
        self._skeletons = []
        self._add_skeleton(self._coord_signature, cube.lazy_data())

        # The nominated axis of concatenation.
        self._axis = None

    @property
    def axis(self):
        """Return the nominated dimension of concatenation."""

        return self._axis

    def concatenate(self):
        """
        Concatenates all the source-cubes registered with the
        :class:`_ProtoCube` over the nominated common dimension.

        Returns:
            The concatenated :class:`iris.cube.Cube`.

        """
        if len(self._skeletons) > 1:
            skeletons = self._skeletons
            dim_ind = self._coord_signature.dim_mapping.index(self.axis)
            order = self._coord_signature.dim_order[dim_ind]
            cube_signature = self._cube_signature

            # Sequence the skeleton segments into the correct order
            # pending concatenation.
            skeletons.sort(
                key=lambda skeleton: skeleton.signature.dim_extents,
                reverse=(order == _DECREASING),
            )

            # Concatenate the new dimension coordinate.
            dim_coords_and_dims = self._build_dim_coordinates()

            # Concatenate the new auxiliary coordinates.
            aux_coords_and_dims = self._build_aux_coordinates()

            # Concatenate the new cell measures
            cell_measures_and_dims = self._build_cell_measures()

            # Concatenate the new ancillary variables
            ancillary_variables_and_dims = self._build_ancillary_variables()

            # Concatenate the new data payload.
            data = self._build_data()

            # Build the new cube.
            kwargs = cube_signature.defn._asdict()
            cube = iris.cube.Cube(
                data,
                dim_coords_and_dims=dim_coords_and_dims,
                aux_coords_and_dims=aux_coords_and_dims,
                cell_measures_and_dims=cell_measures_and_dims,
                ancillary_variables_and_dims=ancillary_variables_and_dims,
                **kwargs,
            )
        else:
            # There are no other source-cubes to concatenate
            # with this proto-cube.
            cube = self._cube

        return cube

    def register(
        self,
        cube,
        axis=None,
        error_on_mismatch=False,
        check_aux_coords=False,
        check_cell_measures=False,
        check_ancils=False,
    ):
        """
        Determine whether the given source-cube is suitable for concatenation
        with this :class:`_ProtoCube`.

        Args:

        * cube:
            The :class:`iris.cube.Cube` source-cube candidate for
            concatenation.

        Kwargs:

        * axis:
            Seed the dimension of concatenation for the :class:`_ProtoCube`
            rather than rely on negotiation with source-cubes.

        * error_on_mismatch:
            If True, raise an informative error if registration fails.

        Returns:
            Boolean.

        """
        # Verify and assert the nominated axis.
        if axis is not None and self.axis is not None and self.axis != axis:
            msg = (
                "Nominated axis [{}] is not equal "
                "to negotiated axis [{}]".format(axis, self.axis)
            )
            raise ValueError(msg)

        # Check for compatible cube signatures.
        cube_signature = _CubeSignature(cube)
        match = self._cube_signature.match(cube_signature, error_on_mismatch)

        # Check for compatible coordinate signatures.
        if match:
            coord_signature = _CoordSignature(cube_signature)
            candidate_axis = self._coord_signature.candidate_axis(
                coord_signature
            )
            match = candidate_axis is not None and (
                candidate_axis == axis or axis is None
            )

        # Check for compatible coordinate extents.
        if match:
            dim_ind = self._coord_signature.dim_mapping.index(candidate_axis)
            match = self._sequence(
                coord_signature.dim_extents[dim_ind], candidate_axis
            )

        # Check for compatible AuxCoords.
        if match:
            if check_aux_coords:
                for coord_a, coord_b in zip(
                    self._cube_signature.aux_coords_and_dims,
                    cube_signature.aux_coords_and_dims,
                ):
                    # AuxCoords that span the candidate axis can differ
                    if (
                        candidate_axis not in coord_a.dims
                        or candidate_axis not in coord_b.dims
                    ):
                        if not coord_a == coord_b:
                            match = False

        # Check for compatible CellMeasures.
        if match:
            if check_cell_measures:
                for coord_a, coord_b in zip(
                    self._cube_signature.cell_measures_and_dims,
                    cube_signature.cell_measures_and_dims,
                ):
                    # CellMeasures that span the candidate axis can differ
                    if (
                        candidate_axis not in coord_a.dims
                        or candidate_axis not in coord_b.dims
                    ):
                        if not coord_a == coord_b:
                            match = False

        # Check for compatible AncillaryVariables.
        if match:
            if check_ancils:
                for coord_a, coord_b in zip(
                    self._cube_signature.ancillary_variables_and_dims,
                    cube_signature.ancillary_variables_and_dims,
                ):
                    # AncillaryVariables that span the candidate axis can differ
                    if (
                        candidate_axis not in coord_a.dims
                        or candidate_axis not in coord_b.dims
                    ):
                        if not coord_a == coord_b:
                            match = False

        if match:
            # Register the cube as a source-cube for this proto-cube.
            self._add_skeleton(coord_signature, cube.lazy_data())
            # Declare the nominated axis of concatenation.
            self._axis = candidate_axis

        if match:
            # If the protocube dimension order is constant (indicating it was
            # created from a cube with a length 1 dimension coordinate) but
            # a subsequently registered cube has a non-constant dimension
            # order we should use that instead of _CONSTANT to make sure all
            # the ordering checks and sorts work as expected.
            dim_ind = self._coord_signature.dim_mapping.index(candidate_axis)
            existing_order = self._coord_signature.dim_order[dim_ind]
            this_order = coord_signature.dim_order[dim_ind]
            if existing_order == _CONSTANT and this_order != _CONSTANT:
                self._coord_signature.dim_order[dim_ind] = this_order

        return match

    def _add_skeleton(self, coord_signature, data):
        """
        Create and add the source-cube skeleton to the
        :class:`_ProtoCube`.

        Args:

        * coord_signature:
            The :class:`_CoordSignature` of the associated
            given source-cube.

        * data:
            The data payload of an associated :class:`iris.cube.Cube`
            source-cube.

        """
        skeleton = _SkeletonCube(coord_signature, data)
        self._skeletons.append(skeleton)

    def _build_aux_coordinates(self):
        """
        Generate the auxiliary coordinates with associated dimension(s)
        mapping for the new concatenated cube.

        Returns:
            A list of auxiliary coordinates and dimension(s) tuple pairs.

        """
        # Setup convenience hooks.
        skeletons = self._skeletons
        cube_signature = self._cube_signature

        aux_coords_and_dims = []

        # Generate all the auxiliary coordinates for the new concatenated cube.
        for i, (coord, dims) in enumerate(cube_signature.aux_coords_and_dims):
            # Check whether the coordinate spans the nominated
            # dimension of concatenation.
            if self.axis in dims:
                # Concatenate the points together.
                dim = dims.index(self.axis)
                points = [
                    skton.signature.aux_coords_and_dims[i].coord.points
                    for skton in skeletons
                ]
                points = np.concatenate(tuple(points), axis=dim)

                # Concatenate the bounds together.
                bnds = None
                if coord.has_bounds():
                    bnds = [
                        skton.signature.aux_coords_and_dims[i].coord.bounds
                        for skton in skeletons
                    ]
                    bnds = np.concatenate(tuple(bnds), axis=dim)

                # Generate the associated coordinate metadata.
                kwargs = cube_signature.aux_metadata[i].defn._asdict()

                # Build the concatenated coordinate.
                if isinstance(coord, iris.coords.AuxCoord):
                    coord = iris.coords.AuxCoord(points, bounds=bnds, **kwargs)
                else:
                    # Attempt to create a DimCoord, otherwise default to
                    # an AuxCoord on failure.
                    try:
                        coord = iris.coords.DimCoord(
                            points, bounds=bnds, **kwargs
                        )
                    except ValueError:
                        # Ensure to remove the "circular" kwarg, which may be
                        # present in the defn of a DimCoord being demoted.
                        _ = kwargs.pop("circular", None)
                        coord = iris.coords.AuxCoord(
                            points, bounds=bnds, **kwargs
                        )

            aux_coords_and_dims.append((coord.copy(), dims))

        # Generate all the scalar coordinates for the new concatenated cube.
        for coord in cube_signature.scalar_coords:
            aux_coords_and_dims.append((coord.copy(), ()))

        return aux_coords_and_dims

    def _build_cell_measures(self):
        """
        Generate the cell measures with associated dimension(s)
        mapping for the new concatenated cube.

        Returns:
            A list of cell measures and dimension(s) tuple pairs.

        """
        # Setup convenience hooks.
        skeletons = self._skeletons
        cube_signature = self._cube_signature

        cell_measures_and_dims = []

        # Generate all the cell measures for the new concatenated cube.
        for i, (cm, dims) in enumerate(cube_signature.cell_measures_and_dims):
            # Check whether the cell measure spans the nominated
            # dimension of concatenation.
            if self.axis in dims:
                # Concatenate the data together.
                dim = dims.index(self.axis)
                data = [
                    skton.signature.cell_measures_and_dims[i].coord.data
                    for skton in skeletons
                ]
                data = np.concatenate(tuple(data), axis=dim)

                # Generate the associated metadata.
                kwargs = cube_signature.cm_metadata[i].defn._asdict()

                # Build the concatenated coordinate.
                cm = iris.coords.CellMeasure(data, **kwargs)

            cell_measures_and_dims.append((cm.copy(), dims))

        return cell_measures_and_dims

    def _build_ancillary_variables(self):
        """
        Generate the ancillary variables with associated dimension(s)
        mapping for the new concatenated cube.

        Returns:
            A list of ancillary variables and dimension(s) tuple pairs.

        """
        # Setup convenience hooks.
        skeletons = self._skeletons
        cube_signature = self._cube_signature

        ancillary_variables_and_dims = []

        # Generate all the ancillary variables for the new concatenated cube.
        for i, (av, dims) in enumerate(
            cube_signature.ancillary_variables_and_dims
        ):
            # Check whether the ancillary variable spans the nominated
            # dimension of concatenation.
            if self.axis in dims:
                # Concatenate the data together.
                dim = dims.index(self.axis)
                data = [
                    skton.signature.ancillary_variables_and_dims[i].coord.data
                    for skton in skeletons
                ]
                data = np.concatenate(tuple(data), axis=dim)

                # Generate the associated metadata.
                kwargs = cube_signature.av_metadata[i].defn._asdict()

                # Build the concatenated coordinate.
                av = iris.coords.AncillaryVariable(data, **kwargs)

            ancillary_variables_and_dims.append((av.copy(), dims))

        return ancillary_variables_and_dims

    def _build_data(self):
        """
        Generate the data payload for the new concatenated cube.

        Returns:
            The concatenated :class:`iris.cube.Cube` data payload.

        """
        skeletons = self._skeletons
        data = [skeleton.data for skeleton in skeletons]

        data = da.concatenate(data, self.axis)

        return data

    def _build_dim_coordinates(self):
        """
        Generate the dimension coordinates with associated dimension
        mapping for the new concatenated cube.

        Return:
            A list of dimension coordinate and dimension tuple pairs.

        """
        # Setup convenience hooks.
        skeletons = self._skeletons
        axis = self.axis
        dim_ind = self._cube_signature.dim_mapping.index(axis)
        metadata = self._cube_signature.dim_metadata[dim_ind]
        defn, circular = metadata.defn, metadata.kwargs["circular"]

        # Concatenate the points together for the nominated dimension.
        points = [
            skeleton.signature.dim_coords[dim_ind].points
            for skeleton in skeletons
        ]
        points = np.concatenate(tuple(points))

        # Concatenate the bounds together for the nominated dimension.
        bounds = None
        if self._cube_signature.dim_coords[dim_ind].has_bounds():
            bounds = [
                skeleton.signature.dim_coords[dim_ind].bounds
                for skeleton in skeletons
            ]
            bounds = np.concatenate(tuple(bounds))

        # Populate the new dimension coordinate with the concatenated
        # points, bounds and associated metadata.
        kwargs = defn._asdict()
        kwargs["circular"] = circular
        dim_coord = iris.coords.DimCoord(points, bounds=bounds, **kwargs)

        # Generate all the dimension coordinates for the new concatenated cube.
        dim_coords_and_dims = []
        for ind, coord in enumerate(self._cube_signature.dim_coords):
            dim = self._cube_signature.dim_mapping[ind]
            if dim == axis:
                dim_coords_and_dims.append((dim_coord, dim))
            else:
                dim_coords_and_dims.append((coord.copy(), dim))

        return dim_coords_and_dims

    def _sequence(self, extent, axis):
        """
        Determine whether the given extent can be sequenced along with
        all the extents of the source-cubes already registered with
        this :class:`_ProtoCube` into non-overlapping segments for the
        given axis.

        Args:

        * extent:
            The :class:`_CoordExtent` of the candidate source-cube.

        * axis:
            The candidate axis of concatenation.

        Returns:
            Boolean.

        """
        result = True

        # Add the new extent to the current extents collection.
        dim_ind = self._coord_signature.dim_mapping.index(axis)
        dim_extents = [
            skeleton.signature.dim_extents[dim_ind]
            for skeleton in self._skeletons
        ]
        dim_extents.append(extent)

        # Sort into the appropriate dimension order.
        order = self._coord_signature.dim_order[dim_ind]
        dim_extents.sort(reverse=(order == _DECREASING))

        # Ensure that the extents don't overlap.
        if len(dim_extents) > 1:
            for i, extent in enumerate(dim_extents[1:]):
                # Check the points - must be strictly monotonic.
                if order == _DECREASING:
                    big = dim_extents[i].points.min
                    small = extent.points.max
                else:
                    small = dim_extents[i].points.max
                    big = extent.points.min

                if small >= big:
                    result = False
                    break

                # Check the bounds - must be strictly monotonic.
                if extent.bounds is not None:
                    if order == _DECREASING:
                        big_0 = dim_extents[i].bounds[0].min
                        big_1 = dim_extents[i].bounds[1].min
                        small_0 = extent.bounds[0].max
                        small_1 = extent.bounds[1].max
                    else:
                        small_0 = dim_extents[i].bounds[0].max
                        small_1 = dim_extents[i].bounds[1].max
                        big_0 = extent.bounds[0].min
                        big_1 = extent.bounds[1].min

                    lower_bound_fail = small_0 >= big_0
                    upper_bound_fail = small_1 >= big_1

                    if lower_bound_fail or upper_bound_fail:
                        result = False
                        break

        return result



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris.analysis.geometry._extract_relevant_cube_slice`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.

import iris.tests as tests
import iris.tests.stock as stock

import shapely.geometry

from iris.analysis.geometry import _extract_relevant_cube_slice


class Test(tests.IrisTest):
    def test_polygon_smaller_than_cube(self):
        cube = stock.lat_lon_cube()
        cube.dim_coords[0].guess_bounds()
        cube.dim_coords[1].guess_bounds()
        geometry = shapely.geometry.box(-0.4, -0.4, 0.4, 0.4)
        actual = _extract_relevant_cube_slice(cube, geometry)
        target = (
            cube[1, 1],
            cube[1, 1].coords(axis="x")[0],
            cube[1, 1].coords(axis="y")[0],
            (1, 1, 1, 1),
        )
        self.assertEqual(target, actual)

    def test_polygon_larger_than_cube(self):
        cube = stock.lat_lon_cube()
        cube.dim_coords[0].guess_bounds()
        cube.dim_coords[1].guess_bounds()
        geometry = shapely.geometry.box(-0.6, -0.6, 0.6, 0.6)
        actual = _extract_relevant_cube_slice(cube, geometry)
        target = (
            cube[:, :3],
            cube[:, :3].coords(axis="x")[0],
            cube[:, :3].coords(axis="y")[0],
            (0, 0, 2, 2),
        )
        self.assertEqual(target, actual)

    def test_polygon_on_cube_boundary(self):
        cube = stock.lat_lon_cube()
        cube.dim_coords[0].guess_bounds()
        cube.dim_coords[1].guess_bounds()
        geometry = shapely.geometry.box(-0.5, -0.5, 0.5, 0.5)
        actual = _extract_relevant_cube_slice(cube, geometry)
        target = (
            cube[1, 1],
            cube[1, 1].coords(axis="x")[0],
            cube[1, 1].coords(axis="y")[0],
            (1, 1, 1, 1),
        )
        self.assertEqual(target, actual)

    def test_rotated_polygon_on_cube_boundary(self):
        cube = stock.lat_lon_cube()
        cube.dim_coords[0].guess_bounds()
        cube.dim_coords[1].guess_bounds()
        geometry = shapely.geometry.Polygon(
            ((0.0, -0.5), (-0.5, 0.0), (0.0, 0.5), (0.5, 0.0))
        )
        actual = _extract_relevant_cube_slice(cube, geometry)
        target = (
            cube[1, 1],
            cube[1, 1].coords(axis="x")[0],
            cube[1, 1].coords(axis="y")[0],
            (1, 1, 1, 1),
        )
        self.assertEqual(target, actual)

    def test_rotated_polygon_larger_than_cube_boundary(self):
        cube = stock.lat_lon_cube()
        cube.dim_coords[0].guess_bounds()
        cube.dim_coords[1].guess_bounds()
        geometry = shapely.geometry.Polygon(
            ((0.0, -0.6), (-0.6, 0.0), (0.0, 0.6), (0.6, 0.0))
        )
        actual = _extract_relevant_cube_slice(cube, geometry)
        target = (
            cube[:, :3],
            cube[:, :3].coords(axis="x")[0],
            cube[:, :3].coords(axis="y")[0],
            (0, 0, 2, 2),
        )
        self.assertEqual(target, actual)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.get_names`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import get_names


class TestGetNames(tests.IrisTest):
    """
    The tests included in this class cover all the variations of possible
    combinations of the following inputs:
    * standard_name = [None, 'projection_y_coordinate', 'latitude_coordinate']
    * long_name = [None, 'lat_long_name']
    * var_name = ['grid_latitude', 'lat_var_name']
    * coord_name = [None, 'latitude']

    standard_name, var_name and coord_name each contain a different valid CF
    standard name so that it is clear which is being used to set the resulting
    standard_name.

    """
    @staticmethod
    def _make_cf_var(standard_name, long_name, cf_name):
        cf_var = mock.Mock(
            cf_name=cf_name,
            standard_name=standard_name,
            long_name=long_name,
            units='degrees',
            dtype=np.float64,
            cell_methods=None,
            cf_group=mock.Mock(global_attributes={}))
        return cf_var

    def check_names(self, inputs, expected):
        # Inputs - attributes on the fake CF Variable. Note: coord_name is
        # optionally set in some pyke rules.
        standard_name, long_name, var_name, coord_name = inputs
        # Expected - The expected names and attributes.
        exp_std_name, exp_long_name, exp_var_name, exp_attributes = expected

        cf_var = self._make_cf_var(standard_name=standard_name,
                                   long_name=long_name, cf_name=var_name)
        attributes = {}
        res_standard_name, res_long_name, res_var_name = get_names(
            cf_var, coord_name, attributes)

        # Check the names and attributes are as expected.
        self.assertEqual(res_standard_name, exp_std_name)
        self.assertEqual(res_long_name, exp_long_name)
        self.assertEqual(res_var_name, exp_var_name)
        self.assertEqual(attributes, exp_attributes)

    def test_var_name_valid(self):
        # Only var_name is set and it is set to a valid standard name.
        inp = (None, None, 'grid_latitude', None)
        exp = ('grid_latitude', None, 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_var_name_valid_coord_name_set(self):
        # var_name is a valid standard name, coord_name is also set.
        inp = (None, None, 'grid_latitude', 'latitude')
        exp = ('latitude', None, 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_var_name_invalid(self):
        # Only var_name is set but it is not a valid standard name.
        inp = (None, None, 'lat_var_name', None)
        exp = (None, None, 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_var_name_invalid_coord_name_set(self):
        # var_name is not a valid standard name, the coord_name is also set.
        inp = (None, None, 'lat_var_name', 'latitude')
        exp = ('latitude', None, 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_long_name_set_var_name_valid(self):
        # long_name is not None, var_name is set to a valid standard name.
        inp = (None, 'lat_long_name', 'grid_latitude', None)
        exp = ('grid_latitude', 'lat_long_name', 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_long_name_set_var_name_valid_coord_name_set(self):
        # long_name is not None, var_name is set to a valid standard name, and
        # coord_name is set.
        inp = (None, 'lat_long_name', 'grid_latitude', 'latitude')
        exp = ('latitude', 'lat_long_name', 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_long_name_set_var_name_invalid(self):
        # long_name is not None, var_name is not set to a valid standard name.
        inp = (None, 'lat_long_name', 'lat_var_name', None)
        exp = (None, 'lat_long_name', 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_long_name_set_var_name_invalid_coord_name_set(self):
        # long_name is not None, var_name is not set to a valid standard name,
        # and coord_name is set.
        inp = (None, 'lat_long_name', 'lat_var_name', 'latitude')
        exp = ('latitude', 'lat_long_name', 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_std_name_valid_var_name_valid(self):
        # standard_name is a valid standard name, var_name is a valid standard
        # name.
        inp = ('projection_y_coordinate', None, 'grid_latitude', None)
        exp = ('projection_y_coordinate', None, 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_std_name_valid_var_name_valid_coord_name_set(self):
        # standard_name is a valid standard name, var_name is a valid standard
        # name, coord_name is set.
        inp = ('projection_y_coordinate', None, 'grid_latitude', 'latitude')
        exp = ('projection_y_coordinate', None, 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_std_name_valid_var_name_invalid(self):
        # standard_name is a valid standard name, var_name is not a valid
        # standard name.
        inp = ('projection_y_coordinate', None, 'lat_var_name', None)
        exp = ('projection_y_coordinate', None, 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_std_name_valid_var_name_invalid_coord_name_set(self):
        # standard_name is a valid standard name, var_name is not a valid
        # standard name, coord_name is set.
        inp = ('projection_y_coordinate', None, 'lat_var_name', 'latitude')
        exp = ('projection_y_coordinate', None, 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_std_name_valid_long_name_set_var_name_valid(self):
        # standard_name is a valid standard name, long_name is not None,
        # var_name is a valid standard name.
        inp = ('projection_y_coordinate', 'lat_long_name', 'grid_latitude',
               None)
        exp = ('projection_y_coordinate', 'lat_long_name', 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_std_name_valid_long_name_set_var_name_valid_coord_name_set(self):
        # standard_name is a valid standard name, long_name is not None,
        # var_name is a valid standard name, coord_name is set.
        inp = ('projection_y_coordinate', 'lat_long_name', 'grid_latitude',
               'latitude')
        exp = ('projection_y_coordinate', 'lat_long_name', 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_std_name_valid_long_name_set_var_name_invalid(self):
        # standard_name is a valid standard name, long_name is not None,
        # var_name is not a valid standard name.
        inp = ('projection_y_coordinate', 'lat_long_name', 'lat_var_name',
               None)
        exp = ('projection_y_coordinate', 'lat_long_name', 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_std_name_valid_long_name_set_var_name_invalid_coord_name_set(
            self):
        # standard_name is a valid standard name, long_name is not None,
        # var_name is not a valid standard name, coord_name is set.
        inp = ('projection_y_coordinate', 'lat_long_name', 'lat_var_name',
               'latitude')
        exp = ('projection_y_coordinate', 'lat_long_name', 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_std_name_invalid_var_name_valid(self):
        # standard_name is not a valid standard name, var_name is a valid
        # standard name.
        inp = ('latitude_coord', None, 'grid_latitude', None)
        exp = ('grid_latitude', None, 'grid_latitude', {})
        self.check_names(inp, exp)

    def test_std_name_invalid_var_name_valid_coord_name_set(self):
        # standard_name is not a valid standard name, var_name is a valid
        # standard name, coord_name is set.
        inp = ('latitude_coord', None, 'grid_latitude', 'latitude')
        exp = ('latitude', None, 'grid_latitude',
               {'invalid_standard_name': 'latitude_coord'})
        self.check_names(inp, exp)

    def test_std_name_invalid_var_name_invalid(self):
        # standard_name is not a valid standard name, var_name is not a valid
        # standard name.
        inp = ('latitude_coord', None, 'lat_var_name', None)
        exp = (None, None, 'lat_var_name', {})
        self.check_names(inp, exp)

    def test_std_name_invalid_var_name_invalid_coord_name_set(self):
        # standard_name is not a valid standard name, var_name is not a valid
        # standard name, coord_name is set.
        inp = ('latitude_coord', None, 'lat_var_name', 'latitude')
        exp = ('latitude', None, 'lat_var_name',
               {'invalid_standard_name': 'latitude_coord'})
        self.check_names(inp, exp)

    def test_std_name_invalid_long_name_set_var_name_valid(self):
        # standard_name is not a valid standard name, long_name is not None
        # var_name is a valid standard name.
        inp = ('latitude_coord', 'lat_long_name', 'grid_latitude', None)
        exp = ('grid_latitude', 'lat_long_name', 'grid_latitude',
               {'invalid_standard_name': 'latitude_coord'})
        self.check_names(inp, exp)

    def test_std_name_invalid_long_name_set_var_name_valid_coord_name_set(
            self):
        # standard_name is not a valid standard name, long_name is not None,
        # var_name is a valid standard name, coord_name is set.
        inp = ('latitude_coord', 'lat_long_name', 'grid_latitude', 'latitude')
        exp = ('latitude', 'lat_long_name', 'grid_latitude',
               {'invalid_standard_name': 'latitude_coord'})
        self.check_names(inp, exp)

    def test_std_name_invalid_long_name_set_var_name_invalid(self):
        # standard_name is not a valid standard name, long_name is not None
        # var_name is not a valid standard name.
        inp = ('latitude_coord', 'lat_long_name', 'lat_var_name', None)
        exp = (None, 'lat_long_name', 'lat_var_name',
               {'invalid_standard_name': 'latitude_coord'})
        self.check_names(inp, exp)

    def test_std_name_invalid_long_name_set_var_name_invalid_coord_name_set(
            self):
        # standard_name is not a valid standard name, long_name is not None,
        # var_name is not a valid standard name, coord_name is set.
        inp = ('latitude_coord', 'lat_long_name', 'lat_var_name', 'latitude')
        exp = ('latitude', 'lat_long_name', 'lat_var_name',
               {'invalid_standard_name': 'latitude_coord'})
        self.check_names(inp, exp)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the file saving mechanism.

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import os
import unittest

import iris
import iris.cube
import iris.util
import iris.fileformats.pp as pp
import iris.fileformats.dot as dot

from io import StringIO

# Make a test skip decorator, for when DOT not available
skip_dotpng = unittest.skipIf(
    not dot.DOT_AVAILABLE,
    'Test(s) require the "dot" executable, which was not found. '
    "Check the dot_path setting in site.cfg.",
)

CHKSUM_ERR = "Mismatch between checksum of iris.save and {}.save."


def save_by_filename(filename1, filename2, cube, saver_fn, iosaver=None):
    """Saves a cube to two different filenames using iris.save and the save method of the object representing the file type directly"""
    # Save from object direct
    saver_fn(cube, filename1)

    # Call save on iris
    iris.save(
        cube, filename2, iosaver
    )  # Optional iris.io.find_saver passed in from test


def save_by_filehandle(
    filehandle1, filehandle2, cube, fn_saver, binary_mode=True
):
    """Saves a cube to two different filehandles using iris.save and the save method of the object representing the file type directly"""
    mode = "wb" if binary_mode else "w"

    # Save from object direct
    with open(filehandle1, mode) as outfile:
        fn_saver(cube, outfile)

    # Call save on iris
    with open(filehandle2, mode) as outfile:
        iris.save(cube, outfile)


@tests.skip_data
class TestSaveMethods(tests.IrisTest):
    """Base class for file saving tests. Loads data and creates/deletes tempfiles"""

    def setUp(self):
        self.cube1 = iris.load_cube(
            tests.get_data_path(("PP", "aPPglob1", "global.pp"))
        )
        self.cube2 = iris.load_cube(
            tests.get_data_path(("PP", "aPPglob1", "global_t_forecast.pp"))
        )
        self.temp_filename1 = iris.util.create_temp_filename(self.ext)
        self.temp_filename2 = iris.util.create_temp_filename(self.ext)

    def tearDown(self):
        for tempfile in (self.temp_filename1, self.temp_filename2):
            try:
                os.remove(tempfile)
            except Exception:
                pass


class TestSavePP(TestSaveMethods):
    """Test saving cubes to PP format"""

    ext = ".pp"

    def test_filename(self):
        # Save using iris.save and pp.save
        save_by_filename(
            self.temp_filename1, self.temp_filename2, self.cube1, pp.save
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

    def test_filehandle(self):
        # Save using iris.save and pp.save
        save_by_filehandle(
            self.temp_filename1,
            self.temp_filename2,
            self.cube1,
            pp.save,
            binary_mode=True,
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

        # Check we can't save when file handle is not binary
        with self.assertRaises(ValueError):
            save_by_filehandle(
                self.temp_filename1,
                self.temp_filename2,
                self.cube1,
                pp.save,
                binary_mode=False,
            )


class TestSaveDot(TestSaveMethods):
    """Test saving cubes to DOT format"""

    ext = ".dot"

    def test_filename(self):
        # Save using iris.save and dot.save
        save_by_filename(
            self.temp_filename1, self.temp_filename2, self.cube1, dot.save
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

    def test_filehandle(self):
        # Save using iris.save and dot.save
        save_by_filehandle(
            self.temp_filename1,
            self.temp_filename2,
            self.cube1,
            dot.save,
            binary_mode=False,
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

        # Check we can't save when file handle is binary
        with self.assertRaises(ValueError):
            save_by_filehandle(
                self.temp_filename1,
                self.temp_filename2,
                self.cube1,
                dot.save,
                binary_mode=True,
            )

    def test_bytesio(self):
        sio = StringIO()

        # Save from dot direct
        dot.save(self.cube1, self.temp_filename1)

        # Call save on iris
        iris.save(self.cube1, sio, iris.io.find_saver(self.ext))

        with open(self.temp_filename1) as infile:
            data = infile.read()

        # Compare files
        self.assertEquals(
            data,
            sio.getvalue(),
            "Mismatch in data when comparing iris bytesio save "
            "and dot.save.",
        )


@skip_dotpng
class TestSavePng(TestSaveMethods):
    """Test saving cubes to png"""

    ext = ".dotpng"

    def test_filename(self):
        # Save using iris.save and dot.save_png
        save_by_filename(
            self.temp_filename1, self.temp_filename2, self.cube1, dot.save_png
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

    def test_filehandle(self):
        # Save using iris.save and dot.save_png
        save_by_filehandle(
            self.temp_filename1,
            self.temp_filename2,
            self.cube1,
            dot.save_png,
            binary_mode=True,
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

        # Check we can't save when file handle is not binary
        with self.assertRaises(ValueError):
            save_by_filehandle(
                self.temp_filename1,
                self.temp_filename2,
                self.cube1,
                dot.save_png,
                binary_mode=False,
            )


class TestSaver(TestSaveMethods):
    """Test saving to Iris when we define the saver type to use"""

    ext = ".spam"

    def test_pp(self):
        # Make our own saver
        pp_saver = iris.io.find_saver("PP")
        save_by_filename(
            self.temp_filename1,
            self.temp_filename2,
            self.cube1,
            pp.save,
            pp_saver,
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

    def test_dot(self):
        # Make our own saver
        dot_saver = iris.io.find_saver("DOT")
        save_by_filename(
            self.temp_filename1,
            self.temp_filename2,
            self.cube1,
            dot.save,
            dot_saver,
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )

    @skip_dotpng
    def test_png(self):
        # Make our own saver
        png_saver = iris.io.find_saver("DOTPNG")
        save_by_filename(
            self.temp_filename1,
            self.temp_filename2,
            self.cube1,
            dot.save_png,
            png_saver,
        )

        # Compare files
        self.assertEqual(
            self.file_checksum(self.temp_filename2),
            self.file_checksum(self.temp_filename1),
            CHKSUM_ERR.format(self.ext),
        )


class TestSaveInvalid(TestSaveMethods):
    """Test iris cannot automatically save to file extensions it does not know about"""

    ext = ".invalid"

    def test_filename(self):
        # Check we can't save a file with an unhandled extension
        with self.assertRaises(ValueError):
            iris.save(self.cube1, self.temp_filename2)


if __name__ == "__main__":
    tests.main()



"""
Oceanographic Profiles and T-S Diagrams
=======================================

This example demonstrates how to plot vertical profiles of different
variables in the same axes, and how to make a scatter plot of two
variables. There is an oceanographic theme but the same techniques are
equally applicable to atmospheric or other kinds of data.

The data used are profiles of potential temperature and salinity in the
Equatorial and South Atlantic, output from an ocean model.

The y-axis of the first plot produced will be automatically inverted due to the
presence of the attribute positive=down on the depth coordinate. This means
depth values intuitively increase downward on the y-axis.

"""

import matplotlib.pyplot as plt

import iris
import iris.iterate
import iris.plot as iplt


def main():
    # Load the gridded temperature and salinity data.
    fname = iris.sample_data_path("atlantic_profiles.nc")
    cubes = iris.load(fname)
    (theta,) = cubes.extract("sea_water_potential_temperature")
    (salinity,) = cubes.extract("sea_water_practical_salinity")

    # Extract profiles of temperature and salinity from a particular point in
    # the southern portion of the domain, and limit the depth of the profile
    # to 1000m.
    lon_cons = iris.Constraint(longitude=330.5)
    lat_cons = iris.Constraint(latitude=lambda l: -10 < l < -9)
    depth_cons = iris.Constraint(depth=lambda d: d <= 1000)
    theta_1000m = theta.extract(depth_cons & lon_cons & lat_cons)
    salinity_1000m = salinity.extract(depth_cons & lon_cons & lat_cons)

    # Plot these profiles on the same set of axes. In each case we call plot
    # with two arguments, the cube followed by the depth coordinate. Putting
    # them in this order places the depth coordinate on the y-axis.
    # The first plot is in the default axes. We'll use the same color for the
    # curve and its axes/tick labels.
    plt.figure(figsize=(5, 6))
    temperature_color = (0.3, 0.4, 0.5)
    ax1 = plt.gca()
    iplt.plot(
        theta_1000m,
        theta_1000m.coord("depth"),
        linewidth=2,
        color=temperature_color,
        alpha=0.75,
    )
    ax1.set_xlabel("Potential Temperature / K", color=temperature_color)
    ax1.set_ylabel("Depth / m")
    for ticklabel in ax1.get_xticklabels():
        ticklabel.set_color(temperature_color)

    # To plot salinity in the same axes we use twiny(). We'll use a different
    # color to identify salinity.
    salinity_color = (0.6, 0.1, 0.15)
    ax2 = plt.gca().twiny()
    iplt.plot(
        salinity_1000m,
        salinity_1000m.coord("depth"),
        linewidth=2,
        color=salinity_color,
        alpha=0.75,
    )
    ax2.set_xlabel("Salinity / PSU", color=salinity_color)
    for ticklabel in ax2.get_xticklabels():
        ticklabel.set_color(salinity_color)
    plt.tight_layout()
    iplt.show()

    # Now plot a T-S diagram using scatter. We'll use all the profiles here,
    # and each point will be coloured according to its depth.
    plt.figure(figsize=(6, 6))
    depth_values = theta.coord("depth").points
    for s, t in iris.iterate.izip(salinity, theta, coords="depth"):
        iplt.scatter(s, t, c=depth_values, marker="+", cmap="RdYlBu_r")
    ax = plt.gca()
    ax.set_xlabel("Salinity / PSU")
    ax.set_ylabel("Potential Temperature / K")
    cb = plt.colorbar(orientation="horizontal")
    cb.set_label("Depth / m")
    plt.tight_layout()
    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function
:func:`iris.fileformats.um._optimal_array_structuring.optimal_array_structure`.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import numpy as np

from iris.fileformats.um._optimal_array_structuring import (
    optimal_array_structure,
)


class Test__optimal_dimensioning_structure:
    pass


class Test_optimal_array_structure(tests.IrisTest):
    def _check_arrays_and_dims(self, result, spec):
        self.assertEqual(set(result.keys()), set(spec.keys()))
        for keyname in spec.keys():
            result_array, result_dims = result[keyname]
            spec_array, spec_dims = spec[keyname]
            self.assertEqual(
                result_dims,
                spec_dims,
                'element dims differ for "{}": '
                "result={!r}, expected {!r}".format(
                    keyname, result_dims, spec_dims
                ),
            )
            self.assertArrayEqual(
                result_array,
                spec_array,
                'element arrays differ for "{}": '
                "result={!r}, expected {!r}".format(
                    keyname, result_array, spec_array
                ),
            )

    def test_none(self):
        with self.assertRaises(IndexError):
            _ = optimal_array_structure([], [])

    def test_one(self):
        # A single value does not make a dimension (no length-1 dims).
        elements = [("a", np.array([1]))]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, ())
        self.assertEqual(primaries, set())
        self.assertEqual(elems_and_dims, {})

    def test_1d(self):
        elements = [("a", np.array([1, 2, 4]))]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (3,))
        self.assertEqual(primaries, set("a"))
        self._check_arrays_and_dims(
            elems_and_dims, {"a": (np.array([1, 2, 4]), (0,))}
        )

    def test_1d_actuals(self):
        # Test use of alternate element values for array construction.
        elements = [("a", np.array([1, 2, 4]))]
        actual_values = [("a", np.array([7, 3, 9]))]
        shape, primaries, elems_and_dims = optimal_array_structure(
            elements, actual_values
        )
        self.assertEqual(shape, (3,))
        self.assertEqual(primaries, set("a"))
        self._check_arrays_and_dims(
            elems_and_dims, {"a": (np.array([7, 3, 9]), (0,))}
        )

    def test_actuals_mismatch_fail(self):
        elements = [("a", np.array([1, 2, 4]))]
        actual_values = [("b", np.array([7, 3, 9]))]
        with self.assertRaisesRegex(ValueError, "Names.* do not match.*"):
            shape, primaries, elems_and_dims = optimal_array_structure(
                elements, actual_values
            )

    def test_2d(self):
        elements = [
            ("a", np.array([2, 2, 2, 3, 3, 3])),
            ("b", np.array([7, 8, 9, 7, 8, 9])),
        ]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (2, 3))
        self.assertEqual(primaries, set(["a", "b"]))
        self._check_arrays_and_dims(
            elems_and_dims,
            {"a": (np.array([2, 3]), (0,)), "b": (np.array([7, 8, 9]), (1,))},
        )

    def test_2d_with_element_values(self):
        # Confirm that elements values are used in the output when supplied.
        elements = [
            ("a", np.array([2, 2, 2, 3, 3, 3])),
            ("b", np.array([7, 8, 9, 7, 8, 9])),
        ]
        elements_values = [
            ("a", np.array([6, 6, 6, 8, 8, 8])),
            ("b", np.array([3, 4, 5, 3, 4, 5])),
        ]
        shape, primaries, elems_and_dims = optimal_array_structure(
            elements, elements_values
        )
        self.assertEqual(shape, (2, 3))
        self.assertEqual(primaries, set(["a", "b"]))
        self._check_arrays_and_dims(
            elems_and_dims,
            {"a": (np.array([6, 8]), (0,)), "b": (np.array([3, 4, 5]), (1,))},
        )

    def test_non_2d(self):
        # An incomplete 2d expansion just becomes 1d
        elements = [
            ("a", np.array([2, 2, 2, 3, 3])),
            ("b", np.array([7, 8, 9, 7, 8])),
        ]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (5,))
        self.assertEqual(primaries, set())
        self._check_arrays_and_dims(
            elems_and_dims,
            {
                "a": (np.array([2, 2, 2, 3, 3]), (0,)),
                "b": (np.array([7, 8, 9, 7, 8]), (0,)),
            },
        )

    def test_degenerate(self):
        # A all-same vector does not appear in the output.
        elements = [("a", np.array([1, 2, 3])), ("b", np.array([4, 4, 4]))]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (3,))
        self.assertEqual(primaries, set(["a"]))
        self._check_arrays_and_dims(
            elems_and_dims, {"a": (np.array([1, 2, 3]), (0,))}
        )

    def test_1d_duplicates(self):
        # When two have the same structure, the first is 'the dimension'.
        elements = [("a", np.array([1, 3, 4])), ("b", np.array([6, 7, 9]))]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (3,))
        self.assertEqual(primaries, set("a"))
        self._check_arrays_and_dims(
            elems_and_dims,
            {
                "a": (np.array([1, 3, 4]), (0,)),
                "b": (np.array([6, 7, 9]), (0,)),
            },
        )

    def test_1d_duplicates_order(self):
        # Same as previous but reverse passed order of elements 'a' and 'b'.
        elements = [("b", np.array([6, 7, 9])), ("a", np.array([1, 3, 4]))]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (3,))
        # The only difference is the one chosen as 'principal'
        self.assertEqual(primaries, set("b"))
        self._check_arrays_and_dims(
            elems_and_dims,
            {
                "a": (np.array([1, 3, 4]), (0,)),
                "b": (np.array([6, 7, 9]), (0,)),
            },
        )

    def test_3_way(self):
        elements = [
            ("t1", np.array([2, 3, 4])),
            ("t2", np.array([4, 5, 6])),
            ("period", np.array([9, 8, 7])),
        ]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (3,))
        self.assertEqual(primaries, set(["t1"]))
        self._check_arrays_and_dims(
            elems_and_dims,
            {
                "t1": (np.array([2, 3, 4]), (0,)),
                "t2": (np.array([4, 5, 6]), (0,)),
                "period": (np.array([9, 8, 7]), (0,)),
            },
        )

    def test_mixed_dims(self):
        elements = [
            ("t1", np.array([1, 1, 11, 11])),
            ("t2", np.array([15, 16, 25, 26])),
            ("ft", np.array([15, 16, 15, 16])),
        ]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (2, 2))
        self.assertEqual(primaries, set(["t1", "ft"]))
        self._check_arrays_and_dims(
            elems_and_dims,
            {
                "t1": (np.array([1, 11]), (0,)),
                "t2": (np.array([[15, 16], [25, 26]]), (0, 1)),
                "ft": (np.array([15, 16]), (1,)),
            },
        )

    def test_missing_dim(self):
        # Case with no dimension element for dimension 1.
        elements = [
            ("t1", np.array([1, 1, 11, 11])),
            ("t2", np.array([15, 16, 25, 26])),
        ]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (4,))
        # The potential 2d nature can not be recognised.
        # 't1' is auxiliary, as it has duplicate values over the dimension.
        self.assertEqual(primaries, set(["t2"]))
        self._check_arrays_and_dims(
            elems_and_dims,
            {
                "t1": (np.array([1, 1, 11, 11]), (0,)),
                "t2": (np.array([15, 16, 25, 26]), (0,)),
            },
        )

    def test_optimal_structure_decision(self):
        # Checks the optimal structure decision logic is working correctly:
        # given the arrays we have here we would expect 'a' to be the primary
        # dimension, as it has higher priority for being supplied first.
        elements = [
            ("a", np.array([1, 1, 1, 2, 2, 2])),
            ("b", np.array([0, 1, 2, 0, 1, 2])),
            ("c", np.array([11, 11, 11, 14, 14, 14])),
            ("d", np.array([10, 10, 10, 10, 10, 10])),
        ]
        shape, primaries, elems_and_dims = optimal_array_structure(elements)
        self.assertEqual(shape, (2, 3))
        self.assertEqual(primaries, set(["a", "b"]))
        self._check_arrays_and_dims(
            elems_and_dims,
            {
                "a": (np.array([1, 2]), (0,)),
                "c": (np.array([11, 14]), (0,)),
                "b": (np.array([0, 1, 2]), (1,)),
            },
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the :func:`iris.experimental.ugrid.ugrid` function.

"""

import iris.tests as tests

import unittest

# Import pyugrid if installed, else fail quietly + disable all the tests.
try:
    import pyugrid
except (ImportError, AttributeError):
    pyugrid = None
skip_pyugrid = unittest.skipIf(
    condition=pyugrid is None,
    reason="Requires pyugrid, which is not available.",
)

import iris.experimental.ugrid


data_path = (
    "NetCDF",
    "ugrid",
)
file21 = "21_triangle_example.nc"
long_name = "volume flux between cells"


@skip_pyugrid
@tests.skip_data
class TestUgrid(tests.IrisTest):
    def test_ugrid(self):
        path = tests.get_data_path(data_path + (file21,))
        cube = iris.experimental.ugrid.ugrid(path, long_name)
        self.assertTrue(hasattr(cube, "mesh"))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Automatic collation of cubes into higher-dimensional cubes.

Typically the cube merge process is handled by
:method:`iris.cube.CubeList.merge`.

"""

from collections import namedtuple, OrderedDict
from copy import deepcopy

import numpy as np

from iris._lazy_data import (
    as_lazy_data,
    as_concrete_data,
    is_lazy_data,
    multidim_lazy_stack,
)
import iris.coords
from iris.common import CoordMetadata, CubeMetadata
import iris.cube
import iris.exceptions
import iris.util


#
# Private namedtuple wrapper classes.
#
class _Template(
    namedtuple("Template", ["dims", "points", "bounds", "kwargs"])
):
    """
    Common framework from which to build a dimension or auxiliary coordinate.

    Args:

    * dims:
        Tuple of the associated :class:`iris.cube.Cube` data dimension/s
        spanned by this coordinate template.

    * points:
        A :mod:`numpy` array representing the coordinate point data. No
        points data is represented by None.

    * bounds:
        A :mod:`numpy` array representing the coordinate bounds data. No
        bounds data is represented by None.

    * kwargs:
        A dictionary of key/value pairs required to create a coordinate.

    """

    __slots__ = ()


class _CoordMetaData(
    namedtuple("CoordMetaData", ["points_dtype", "bounds_dtype", "kwargs"])
):
    """
    Bespoke metadata required to build a dimension or auxiliary coordinate.

    Args:

    * points_dtype:
        The points data :class:`numpy.dtype` of an associated coordinate.
        None otherwise.

    * bounds_dtype:
        The bounds data :class:`numpy.dtype` of an associated coordinate.
        None otherwise.

    * kwargs:
        A dictionary of key/value pairs required to create a coordinate.

    """

    __slots__ = ()


class _CoordAndDims(namedtuple("CoordAndDims", ["coord", "dims"])):
    """
    Container for a coordinate and the associated data dimension/s
    spanned over a :class:`iris.cube.Cube`.

    Args:

    * coord:
        A :class:`iris.coords.DimCoord` or :class:`iris.coords.AuxCoord`
        coordinate instance.

    * dims:
        A tuple of the data dimension/s spanned by the coordinate.

    """

    __slots__ = ()


class _ScalarCoordPayload(
    namedtuple("ScalarCoordPayload", ["defns", "values", "metadata"])
):
    """
    Container for all scalar coordinate data and metadata represented
    within a :class:`iris.cube.Cube`.

    All scalar coordinate related data is sorted into ascending order
    of the associated coordinate definition.

    Args:

    * defns:
        A list of scalar coordinate metadata :class:`iris.common.CoordMetadata`
        belonging to a :class:`iris.cube.Cube`.

    * values:
        A list of scalar coordinate values belonging to a
        :class:`iris.cube.Cube`.  Each scalar coordinate value is
        typically an :class:`iris.coords.Cell`.

    * metadata:
        A list of :class:`_CoordMetaData` instances belonging to a
        :class:`iris.cube.Cube`.

    """

    __slots__ = ()


class _VectorCoordPayload(
    namedtuple(
        "VectorCoordPayload", ["dim_coords_and_dims", "aux_coords_and_dims"]
    )
):
    """
    Container for all vector coordinate data and metadata represented
    within a :class:`iris.cube.Cube`.

    Args:

    * dim_coords_and_dims:
        A list of :class:`_CoordAndDim` instances containing non-scalar
        (i.e. multi-valued) :class:`iris.coords.DimCoord` instances and
        the associated data dimension spanned by them for a
        :class:`iris.cube.Cube`.

    * aux_coords_and_dims:
        A list of :class:`_CoordAndDim` instances containing non-scalar
        (i.e. multi-valued) :class:`iris.coords.DimCoord` and/or
        :class:`iris.coords.AuxCoord` instances and the associated data
        dimension/s spanned by them for a :class:`iris.cube.Cube`.

    """

    __slots__ = ()


class _CoordPayload(
    namedtuple("CoordPayload", ["scalar", "vector", "factory_defns"])
):
    """
    Container for all the scalar and vector coordinate data and
    metadata, and auxiliary coordinate factories represented within a
    :class:`iris.cube.Cube`.

    All scalar coordinate and factory related data is sorted into
    ascending order of the associated coordinate definition.

    Args:

    * scalar:
        A :class:`_ScalarCoordPayload` instance.

    * vector:
        A :class:`_VectorCoordPayload` instance.

    * factory_defns:
        A list of :class:`_FactoryDefn` instances.

    """

    __slots__ = ()

    def as_signature(self):
        """Construct and return a :class:`_CoordSignature` from the payload."""

        return _CoordSignature(
            self.scalar.defns,
            self.vector.dim_coords_and_dims,
            self.vector.aux_coords_and_dims,
            self.factory_defns,
        )

    @staticmethod
    def _coords_msgs(msgs, coord_group, defns_a, defns_b):
        if defns_a != defns_b:
            # Get a new list so we can modify it
            defns_b = list(defns_b)
            diff_defns = []
            for defn_a in defns_a:
                try:
                    defns_b.remove(defn_a)
                except ValueError:
                    diff_defns.append(defn_a)
            diff_defns.extend(defns_b)
            if diff_defns:
                names = sorted(set(defn.name() for defn in diff_defns))
                msgs.append(
                    "Coordinates in {} differ: {}.".format(
                        coord_group, ", ".join(names)
                    )
                )
            else:
                msgs.append(
                    "Coordinates in {} differ by dtype or class"
                    " (i.e. DimCoord vs AuxCoord).".format(coord_group)
                )

    def match_signature(self, signature, error_on_mismatch):
        """
        Return whether this _CoordPayload matches the corresponding
        aspects of a _CoordSignature.

        Args:

        * signature (_CoordSignature):
            The _CoordSignature to compare against.

        * error_on_mismatch (bool):
            If True, raise an Exception with detailed explanation.

        Returns:
           Boolean. True if and only if this _CoordPayload matches
           the corresponding aspects `other`.

        """

        def unzip(coords_and_dims):
            if coords_and_dims:
                coords, dims = zip(*coords_and_dims)
            else:
                coords, dims = [], []
            return coords, dims

        def dims_msgs(msgs, coord_group, dimlists_a, dimlists_b):
            if dimlists_a != dimlists_b:
                msgs.append(
                    "Coordinate-to-dimension mapping differs for {}.".format(
                        coord_group
                    )
                )

        msgs = []
        self._coords_msgs(
            msgs,
            "cube.aux_coords (scalar)",
            self.scalar.defns,
            signature.scalar_defns,
        )

        coord_group = "cube.dim_coords"
        self_coords, self_dims = unzip(self.vector.dim_coords_and_dims)
        other_coords, other_dims = unzip(signature.vector_dim_coords_and_dims)
        self._coords_msgs(msgs, coord_group, self_coords, other_coords)
        dims_msgs(msgs, coord_group, self_dims, other_dims)

        coord_group = "cube.aux_coords (non-scalar)"
        self_coords, self_dims = unzip(self.vector.aux_coords_and_dims)
        other_coords, other_dims = unzip(signature.vector_aux_coords_and_dims)
        self._coords_msgs(msgs, coord_group, self_coords, other_coords)
        dims_msgs(msgs, coord_group, self_dims, other_dims)

        if self.factory_defns != signature.factory_defns:
            msgs.append("cube.aux_factories() differ")

        match = not bool(msgs)
        if error_on_mismatch and not match:
            raise iris.exceptions.MergeError(msgs)
        return match


class _CoordSignature(
    namedtuple(
        "CoordSignature",
        [
            "scalar_defns",
            "vector_dim_coords_and_dims",
            "vector_aux_coords_and_dims",
            "factory_defns",
        ],
    )
):
    """
    Criterion for identifying a specific type of :class:`iris.cube.Cube`
    based on its scalar and vector coorinate data and metadata, and
    auxiliary coordinate factories.

    Args:

    * scalar_defns:
        A list of scalar coordinate definitions sorted into ascending order.

    * vector_dim_coords_and_dims:
        A list of :class:`_CoordAndDim` instances containing non-scalar
        (i.e. multi-valued) :class:`iris.coords.DimCoord` instances and
        the associated data dimension spanned by them for a
        :class:`iris.cube.Cube`.

    * vector_aux_coords_and_dims:
        A list of :class:`_CoordAndDim` instances containing non-scalar
        (i.e. multi-valued) :class:`iris.coords.DimCoord` and/or
        :class:`iris.coords.AuxCoord` instances and the associated data
        dimension/s spanned by them for a :class:`iris.cube.Cube`.

    * factory_defns:
        A list of :class:`_FactoryDefn` instances.

    """

    __slots__ = ()


class _CubeSignature(
    namedtuple(
        "CubeSignature",
        [
            "defn",
            "data_shape",
            "data_type",
            "cell_measures_and_dims",
            "ancillary_variables_and_dims",
        ],
    )
):
    """
    Criterion for identifying a specific type of :class:`iris.cube.Cube`
    based on its metadata.

    Args:

    * defn:
        A cube definition tuple.

    * data_shape:
        The data payload shape of a :class:`iris.cube.Cube`.

    * data_type:
        The data payload :class:`numpy.dtype` of a :class:`iris.cube.Cube`.

    * cell_measures_and_dims:
        A list of cell_measures and dims for the cube.

    * ancillary_variables_and_dims:
        A list of ancillary variables and dims for the cube.

    """

    __slots__ = ()

    def _defn_msgs(self, other_defn):
        msgs = []
        self_defn = self.defn
        if self_defn.standard_name != other_defn.standard_name:
            msgs.append(
                "cube.standard_name differs: {!r} != {!r}".format(
                    self_defn.standard_name, other_defn.standard_name
                )
            )
        if self_defn.long_name != other_defn.long_name:
            msgs.append(
                "cube.long_name differs: {!r} != {!r}".format(
                    self_defn.long_name, other_defn.long_name
                )
            )
        if self_defn.var_name != other_defn.var_name:
            msgs.append(
                "cube.var_name differs: {!r} != {!r}".format(
                    self_defn.var_name, other_defn.var_name
                )
            )
        if self_defn.units != other_defn.units:
            msgs.append(
                "cube.units differs: {!r} != {!r}".format(
                    self_defn.units, other_defn.units
                )
            )
        if self_defn.attributes != other_defn.attributes:
            diff_keys = set(self_defn.attributes.keys()) ^ set(
                other_defn.attributes.keys()
            )
            if diff_keys:
                msgs.append(
                    "cube.attributes keys differ: "
                    + ", ".join(repr(key) for key in diff_keys)
                )
            else:
                diff_attrs = [
                    repr(key)
                    for key in self_defn.attributes
                    if np.all(
                        self_defn.attributes[key] != other_defn.attributes[key]
                    )
                ]
                diff_attrs = ", ".join(diff_attrs)
                msgs.append(
                    "cube.attributes values differ for keys: {}".format(
                        diff_attrs
                    )
                )
        if self_defn.cell_methods != other_defn.cell_methods:
            msgs.append("cube.cell_methods differ")
        return msgs

    def match(self, other, error_on_mismatch):
        """
        Return whether this _CubeSignature equals another.

        This is the first step to determine if two "cubes" (either a
        real Cube or a ProtoCube) can be merged, by considering:
            - standard_name, long_name, var_name
            - units
            - attributes
            - cell_methods
            - shape, dtype

        Args:

        * other (_CubeSignature):
            The _CubeSignature to compare against.

        * error_on_mismatch (bool):
            If True, raise a :class:`~iris.exceptions.MergeException`
            with a detailed explanation if the two do not match.

        Returns:
           Boolean. True if and only if this _CubeSignature matches `other`.

        """
        msgs = self._defn_msgs(other.defn)
        if self.data_shape != other.data_shape:
            msg = "cube.shape differs: {} != {}"
            msgs.append(msg.format(self.data_shape, other.data_shape))
        if self.data_type != other.data_type:
            msg = "cube data dtype differs: {} != {}"
            msgs.append(msg.format(self.data_type, other.data_type))
        # Both cell_measures_and_dims and ancillary_variables_and_dims are
        # ordered by the same method, it is therefore not possible for a
        # mismatch to be caused by a difference in order.
        if self.cell_measures_and_dims != other.cell_measures_and_dims:
            msgs.append("cube.cell_measures differ")
        if (
            self.ancillary_variables_and_dims
            != other.ancillary_variables_and_dims
        ):
            msgs.append("cube.ancillary_variables differ")

        match = not bool(msgs)
        if error_on_mismatch and not match:
            raise iris.exceptions.MergeError(msgs)
        return match


class _Skeleton(namedtuple("Skeleton", ["scalar_values", "data"])):
    """
    Basis of a source-cube, containing the associated scalar coordinate values
    and data payload of a :class:`iris.cube.Cube`.

    Args:

    * scalar_values:
        A list of scalar coordinate values belonging to a
        :class:`iris.cube.Cube` sorted into ascending order of the
        associated coordinate definition. Each scalar coordinate value
        is typically an :class:`iris.coords.Cell`.

    * data:
        The data payload of a :class:`iris.cube.Cube`.

    """

    __slots__ = ()


class _FactoryDefn(namedtuple("_FactoryDefn", ["class_", "dependency_defns"])):
    """
    The information required to identify and rebuild a single AuxCoordFactory.

    Args:

    * class_:
        The class of the AuxCoordFactory.

    * dependency_defns:
        A list of pairs, where each pair contains a dependency key and its
        corresponding coordinate definition. Sorted on dependency key.

    """

    __slots__ = ()


class _Relation(namedtuple("Relation", ["separable", "inseparable"])):
    """
    Categorisation of the candidate dimensions belonging to a
    :class:`ProtoCube` into separable 'independent' dimensions, and
    inseparable dependent dimensions.

    Args:

    * separable:
        A set of independent candidate dimension names.

    * inseperable:
        A set of dependent candidate dimension names.

    """

    __slots__ = ()


_COMBINATION_JOIN = "-"


def _is_combination(name):
    """
    Determine whether the candidate dimension is an 'invented' combination
    of candidate dimensions.

    Args:

    * name:
        The candidate dimension.

    Returns:
        Boolean.

    """
    return _COMBINATION_JOIN in str(name)


def build_indexes(positions):
    """
    Construct a mapping for each candidate dimension that maps for each
    of its scalar values the set of values for each of the other candidate
    dimensions.

    For example:

        >>> from iris._merge import build_indexes
        >>> positions = [{'a': 0, 'b': 10, 'c': 100},
        ...              {'a': 1, 'b': 10, 'c': 200},
        ...              {'a': 2, 'b': 20, 'c': 300}]
        ...
        >>> indexes = build_indexes(positions)
        >>> for k in sorted(indexes):
        ...     print('%r:' % k)
        ...     for kk in sorted(indexes[k]):
        ...         print('\t%r:' % kk, end=' ')
        ...         for kkk in sorted(indexes[k][kk]):
        ...             print('%r: %r' % (kkk, indexes[k][kk][kkk]), end=' ')
        ...         print()
        ...
        'a':
             0: 'b': set([10]) 'c': set([100])
             1: 'b': set([10]) 'c': set([200])
             2: 'b': set([20]) 'c': set([300])
        'b':
             10: 'a': set([0, 1]) 'c': set([200, 100])
             20: 'a': set([2]) 'c': set([300])
        'c':
             100: 'a': set([0]) 'b': set([10])
             200: 'a': set([1]) 'b': set([10])
             300: 'a': set([2]) 'b': set([20])

    Args:

    * positions:
        A list containing a dictionary of candidate dimension key to
        scalar value pairs for each source-cube.

    Returns:
        The cross-reference dictionary for each candidate dimension.

    """
    names = positions[0].keys()
    scalar_index_by_name = {name: {} for name in names}

    for position in positions:
        for name, value in position.items():
            name_index_by_scalar = scalar_index_by_name[name]

            if value in name_index_by_scalar:
                value_index_by_name = name_index_by_scalar[value]
                for other_name in names:
                    if other_name != name:
                        value_index_by_name[other_name].add(
                            position[other_name]
                        )
            else:
                name_index_by_scalar[value] = {
                    other_name: set((position[other_name],))
                    for other_name in names
                    if other_name != name
                }

    return scalar_index_by_name


def _separable_pair(name, index):
    """
    Determine whether the candidate dimension is separable.

    A candidate dimension X and Y are separable if each scalar
    value of X maps to the same set of scalar values of Y.

    Args:

    * name1:
        The first candidate dimension to be compared.

    * name2:
        The second candidate dimension to be compared.

    * index:
        The cross-reference dictionary for the first candidate
        dimension.

    Returns:
        Boolean.

    """
    items = iter(index.values())
    reference = next(items)[name]

    return all([item[name] == reference for item in items])


def _separable(name, indexes):
    """
    Determine the candidate dimensions that are separable and
    inseparable relative to the provided candidate dimension.

    A candidate dimension X and Y are separable if each scalar
    value of X maps to the same set of scalar values of Y.

    Args:

    * name:
        The candidate dimension that requires its separable and
        inseparable relationship to be determined.

    * indexes:
        The cross-reference dictionary for each candidate dimension.

    Returns:
        A tuple containing the set of separable and inseparable
        candidate dimensions.

    """
    separable = set()
    inseparable = set()

    for target_name in indexes:
        if name != target_name:
            if _separable_pair(target_name, indexes[name]):
                separable.add(target_name)
            else:
                inseparable.add(target_name)

    return _Relation(separable, inseparable)


def derive_relation_matrix(indexes):
    """
    Construct a mapping for each candidate dimension that specifies
    which of the other candidate dimensions are separable or inseparable.

    A candidate dimension X and Y are separable if each scalar value of
    X maps to the same set of scalar values of Y.

    Also see :func:`iris._merge.build_indexes`.

    For example:

        >>> from iris._merge import build_indexes, derive_relation_matrix
        >>> positions = [{'a': 0, 'b': 10, 'c': 100},
        ...              {'a': 1, 'b': 10, 'c': 200},
        ...              {'a': 2, 'b': 20, 'c': 300}]
        ...
        >>> indexes = build_indexes(positions)
        >>> matrix = derive_relation_matrix(indexes)
        >>> for k, v in matrix.iteritems():
        ...     print('%r: %r' % (k, v))
        ...
        'a': Relation(separable=set([]), inseparable=set(['c', 'b']))
        'c': Relation(separable=set([]), inseparable=set(['a', 'b']))
        'b': Relation(separable=set([]), inseparable=set(['a', 'c']))

    Args:

    * indexes:
        The cross-reference dictionary for each candidate dimension.

    Returns:
        The relation dictionary for each candidate dimension.

    """
    # TODO: This takes twice as long as it could do because it doesn't
    # account for the symmetric nature of the relationship.
    relation_matrix = {name: _separable(name, indexes) for name in indexes}

    return relation_matrix


def derive_groups(relation_matrix):
    """
    Determine all related (chained) groups of inseparable candidate dimensions.

    If candidate dimension A is inseparable for B and C, and B is inseparable
    from D, and E is inseparable from F. Then the groups are ABCD and EF.

    Args:

    * relation_matrix:
        The relation dictionary for each candidate dimension.

    Returns:
        A list of all related (chained) inseparable candidate dimensions.

    """
    names = set(relation_matrix)
    groups = []

    while names:
        name = names.pop()
        group = set([name])
        to_follow = set([name])

        while to_follow:
            name = to_follow.pop()
            new = relation_matrix[name].inseparable - group
            group.update(new)
            to_follow.update(new)
            names -= new

        groups.append(group)

    return groups


def _derive_separable_group(relation_matrix, group):
    """
    Determine which candidate dimensions in the group are separable.

    Args:

    * relation_matrix:
        The relation dictionary for each candidate dimension.

    * group:
        A set of related (chained) inseparable candidate dimensions.

    Returns:
        The set of candidate dimensions within the group that are
        separable.

    """
    result = set()

    for name in group:
        if relation_matrix[name].separable & group:
            result.add(name)

    return result


def _is_dependent(dependent, independent, positions, function_mapping=None):
    """
    Determine whether there exists a one-to-one functional relationship
    between the independent candidate dimension/s and the dependent
    candidate dimension.

    Args:

    * dependent:
        A candidate dimension that requires to be functionally
        dependent on all the independent candidate dimensions.

    * independent:
        A list of candidate dimension/s that require to act as the independent
        variables in a functional relationship.

    * positions:
        A list containing a dictionary of candidate dimension key to
        scalar value pairs for each source-cube.

    Kwargs:

    * function_mapping:
        A dictionary that enumerates a valid functional relationship
        between the dependent candidate dimension and the independent
        candidate dimension/s.

    Returns:
        Boolean.

    """
    valid = True
    relation = {}

    if isinstance(function_mapping, dict):
        relation = function_mapping

    for position in positions:
        item = tuple([position[name] for name in independent])

        if item in relation:
            if position[dependent] != relation[item]:
                valid = False
                break
        else:
            relation[item] = position[dependent]

    return valid


def _derive_consistent_groups(relation_matrix, separable_group):
    """
    Determine the largest combinations of candidate dimensions within the
    separable group that are self consistently separable from one another.

    If the candidate dimension A is separable from the candidate dimensions
    B and C. Then the candidate dimension group ABC is a separable consistent
    group if B is separable from A and C, and C is separable from A and B.

    Args:

    * relation_matrix:
        The relation dictionary for each candidate dimension.

    * separable_group:
        The set of candidate dimensions that are separable.

    Returns:
        A list of candidate dimension groups that are consistently separable.

    """
    result = []

    for name in separable_group:
        name_separable_group = (
            relation_matrix[name].separable & separable_group
        )
        candidate = list(name_separable_group) + [name]
        valid = True

        for _ in range(len(name_separable_group)):
            candidate_separable_group = set(candidate[1:])

            if (
                candidate_separable_group
                & (relation_matrix[candidate[0]].separable & separable_group)
                != candidate_separable_group
            ):
                valid = False
                break

            candidate.append(candidate.pop(0))

        if valid:
            result.append(candidate)

    return result


def _build_separable_group(
    space, group, separable_consistent_groups, positions, function_matrix
):
    """
    Update the space with the first separable consistent group that
    satisfies a valid functional relationship with all other candidate
    dimensions in the group.

    For example, the group ABCD and separable consistent group CD,
    if A = f(C, D) and B = f(C, D) then update the space with
    "A: (C, D), B: (C, D), C: None, D: None". Where "A: (C, D)" means
    that candidate dimension A is dependent on candidate dimensions C
    and D, and "C: None" means that this candidate dimension is
    independent.

    Args:

    * space:
        A dictionary defining for each candidate dimension its
        dependency on any other candidate dimensions within the space.

    * group:
        A set of related (chained) inseparable candidate dimensions.

    * separable_consistent_groups:
        A list of candidate dimension groups that are consistently separable.

    * positions:
        A list containing a dictionary of candidate dimension key to
        scalar value pairs for each source-cube.

    * function_matrix:
        The function mapping dictionary for each candidate dimension that
        participates in a functional relationship.

    Returns:
        Boolean.

    """
    valid = False

    for independent in sorted(separable_consistent_groups):
        dependent = list(group - set(independent))
        dependent_function_matrix = {}

        for name in dependent:
            function_mapping = {}
            valid = _is_dependent(
                name, independent, positions, function_mapping
            )

            if not valid:
                break

            dependent_function_matrix[name] = function_mapping

        if valid:
            break

    if function_matrix is not None:
        function_matrix.update(dependent_function_matrix)

    if valid:
        space.update({name: None for name in independent})
        space.update({name: tuple(independent) for name in dependent})

    return valid


def _build_inseparable_group(space, group, positions, function_matrix):
    """
    Update the space with the first valid scalar functional relationship
    between a candidate dimension within the group and all other
    candidate dimensions.

    For example, with the inseparable group ABCD, a valid scalar
    relationship B = f(A), C = f(A) and D = f(A) results in a space with
    "A: None, B: (A,), C: (A,), D: (A,)".
    Where "A: None" means that this candidate dimension is independent,
    and "B: (A,)" means that candidate dimension B is dependent on
    candidate dimension A.

    The scalar relationship must exist between one candidate dimension
    and all others in the group, as the group is considered inseparable
    in this context.

    Args:

    * space:
        A dictionary defining for each candidate dimension its dependency on
        any other candidate dimensions within the space.

    * group:
        A set of related (chained) inseparable candidate dimensions.

    * positions:
        A list containing a dictionary of candidate dimension key to
        scalar value pairs for each source-cube.

    * function_matrix:
        The function mapping dictionary for each candidate dimension that
        participates in a functional relationship.

    Returns:
        Boolean.

    """
    scalar = False

    for name in sorted(group):
        independent = set([name])
        dependent = set(group) - independent
        valid = False
        dependent_function_matrix = {}

        for name in dependent:
            function_mapping = {}
            valid = _is_dependent(
                name, independent, positions, function_mapping
            )

            if not valid:
                break

            dependent_function_matrix[name] = function_mapping

        if valid:
            scalar = True

            if function_matrix is not None:
                function_matrix.update(dependent_function_matrix)

            space.update({name: None for name in independent})
            space.update({name: tuple(independent) for name in dependent})
            break

    return scalar


def _build_combination_group(space, group, positions, function_matrix):
    """
    Update the space with the new combined or invented dimension
    that each member of this inseparable group depends on.

    As no functional relationship between members of the group can be
    determined, the new combination dimension will not have a dimension
    coordinate associated with it. Rather, it is simply an enumeration
    of the group members for each of the positions (source-cubes).

    Args:

    * space:
        A dictionary defining for each candidate dimension its dependency on
        any other candidate dimensions within the space.

    * group:
        A set of related (chained) inseparable candidate dimensions.

    * positions:
        A list containing a dictionary of candidate dimension key to
        scalar value pairs for each source-cube.

    * function_matrix:
        The function mapping dictionary for each candidate dimension that
        participates in a functional relationship.

    Returns:
        None.

    """
    combination = _COMBINATION_JOIN.join(sorted(map(str, group)))
    space.update({name: None for name in (combination,)})
    space.update({name: (combination,) for name in group})
    members = combination.split(_COMBINATION_JOIN)

    # Populate the function matrix for each member of the group.
    for name in group:
        function_matrix[name] = {}

    for position in positions:
        # Note, the cell double-tuple! This ensures that the cell value for
        # each member of the group is kept bound together as one key.
        cell = (
            tuple(
                [
                    position[int(member) if member.isdigit() else member]
                    for member in members
                ]
            ),
        )
        for name in group:
            function_matrix[name][cell] = position[name]


def derive_space(groups, relation_matrix, positions, function_matrix=None):
    """
    Determine the relationship between all the candidate dimensions.

    Args:
      * groups:
          A list of all related (chained) inseparable candidate dimensions.

      * relation_matrix:
          The relation dictionary for each candidate dimension.

      * positions:
          A list containing a dictionary of candidate dimension key to
          scalar value pairs for each source-cube.

    Kwargs:
      * function_matrix:
          The function mapping dictionary for each candidate dimension that
          participates in a functional relationship.

    Returns:
        A space dictionary describing the relationship between each
        candidate dimension.

    """
    space = {}

    for group in groups:
        separable_group = _derive_separable_group(relation_matrix, group)

        if len(group) == 1 and not separable_group:
            # This single candidate dimension is separable from all other
            # candidate dimensions in the group, therefore it is a genuine
            # dimension of the space.
            space.update({name: None for name in group})
        elif separable_group:
            # Determine the largest combination of the candidate dimensions
            # in the separable group that are consistently separable.
            consistent_groups = _derive_consistent_groups(
                relation_matrix, separable_group
            )
            if not _build_separable_group(
                space, group, consistent_groups, positions, function_matrix
            ):
                # There is no relationship between any of the candidate
                # dimensions in the separable group, so merge them together
                # into a new combined dimension of the space.
                _build_combination_group(
                    space, group, positions, function_matrix
                )
        else:
            # Determine whether there is a scalar relationship between one of
            # the candidate dimensions and each of the other candidate
            # dimensions in this inseparable group.
            if not _build_inseparable_group(
                space, group, positions, function_matrix
            ):
                # There is no relationship between any of the candidate
                # dimensions in this inseparable group, so merge them together
                # into a new combined dimension of the space.
                _build_combination_group(
                    space, group, positions, function_matrix
                )

    return space


class ProtoCube:
    """
    Framework for merging source-cubes into one or more higher
    dimensional cubes.

    """

    def __init__(self, cube):
        """
        Create a new ProtoCube from the given cube and record the cube
        as a source-cube.

        """

        # Default hint ordering for candidate dimension coordinates.
        self._hints = [
            "time",
            "forecast_reference_time",
            "forecast_period",
            "model_level_number",
        ]

        # The proto-cube source.
        self._source = cube

        # The cube signature is metadata that defines this ProtoCube.
        self._cube_signature = self._build_signature(cube)

        # Extract the scalar and vector coordinate data and metadata
        # from the cube.
        coord_payload = self._extract_coord_payload(cube)

        # The coordinate signature defines the scalar and vector
        # coordinates of this ProtoCube.
        self._coord_signature = coord_payload.as_signature()
        self._coord_metadata = coord_payload.scalar.metadata

        # The list of stripped-down source-cubes relevant to this ProtoCube.
        self._skeletons = []
        self._add_cube(cube, coord_payload)

        # Proto-coordinates constructed from merged scalars.
        self._dim_templates = []
        self._aux_templates = []

        # During the merge this will contain the complete, merged shape
        # of a result cube.
        # E.g. Merging three (72, 96) cubes would give:
        #      self._shape = (3, 72, 96).
        self._shape = []
        # During the merge this will contain the shape of the "stack"
        # of cubes used to create a single result cube.
        # E.g. Merging three (72, 96) cubes would give:
        #      self._stack_shape = (3,)
        self._stack_shape = []

        self._nd_names = []
        self._cache_by_name = {}
        self._dim_coords_and_dims = []
        self._aux_coords_and_dims = []

        # Dims offset by merged space higher dimensionality.
        self._vector_dim_coords_dims = []
        self._vector_aux_coords_dims = []

        # cell measures and ancillary variables are not merge candidates
        # they are checked and preserved through merge
        self._cell_measures_and_dims = cube._cell_measures_and_dims
        self._ancillary_variables_and_dims = cube._ancillary_variables_and_dims

    def _report_duplicate(self, nd_indexes, group_by_nd_index):
        # Find the first offending source-cube with duplicate metadata.
        index = [
            group_by_nd_index[nd_index][1]
            for nd_index in nd_indexes
            if len(group_by_nd_index[nd_index]) > 1
        ][0]
        name = self._cube_signature.defn.name()
        scalars = []
        for defn, value in zip(
            self._coord_signature.scalar_defns,
            self._skeletons[index].scalar_values,
        ):
            scalars.append("%s=%r" % (defn.name(), value))
        msg = "Duplicate %r cube, with scalar coordinates %s"
        msg = msg % (name, ", ".join(scalars))
        raise iris.exceptions.DuplicateDataError(msg)

    def merge(self, unique=True):
        """
        Returns the list of cubes resulting from merging the registered
        source-cubes.

        Kwargs:

        * unique:
            If True, raises `iris.exceptions.DuplicateDataError` if
            duplicate cubes are detected.

        Returns:
            A :class:`iris.cube.CubeList` of merged cubes.

        """
        positions = [
            {i: v for i, v in enumerate(skeleton.scalar_values)}
            for skeleton in self._skeletons
        ]
        indexes = build_indexes(positions)
        relation_matrix = derive_relation_matrix(indexes)
        groups = derive_groups(relation_matrix)

        function_matrix = {}
        space = derive_space(
            groups, relation_matrix, positions, function_matrix=function_matrix
        )
        self._define_space(space, positions, indexes, function_matrix)
        self._build_coordinates()

        # All the final, merged cubes will end up here.
        merged_cubes = iris.cube.CubeList()

        # Collate source-cubes by the nd-index.
        group_by_nd_index = {}
        for index, position in enumerate(positions):
            group = group_by_nd_index.setdefault(self._nd_index(position), [])
            group.append(index)

        # Determine the largest group of source-cubes that want to occupy
        # the same nd-index in the final merged cube.
        group_depth = max([len(group) for group in group_by_nd_index.values()])
        nd_indexes = sorted(group_by_nd_index.keys())

        # Check for unique data.
        if unique and group_depth > 1:
            self._report_duplicate(nd_indexes, group_by_nd_index)

        # Generate group-depth merged cubes from the source-cubes.
        for level in range(group_depth):
            # Track the largest dtype of the data to be merged.
            # Unfortunately, da.stack() is not symmetric with regards
            # to dtypes. So stacking float + int yields a float, but
            # stacking an int + float yields an int! We need to ensure
            # that the largest dtype prevails i.e. float, in order to
            # support the masked case for dask.
            # Reference https://github.com/dask/dask/issues/2273.
            dtype = None
            # Stack up all the data from all of the relevant source
            # cubes in a single dask "stacked" array.
            # If it turns out that all the source cubes already had
            # their data loaded then at the end we convert the stack back
            # into a plain numpy array.
            stack = np.empty(self._stack_shape, "object")
            all_have_data = True
            for nd_index in nd_indexes:
                # Get the data of the current existing or last known
                # good source-cube
                group = group_by_nd_index[nd_index]
                offset = min(level, len(group) - 1)
                data = self._skeletons[group[offset]].data
                # Ensure the data is represented as a dask array and
                # slot that array into the stack.
                if is_lazy_data(data):
                    all_have_data = False
                else:
                    data = as_lazy_data(data)
                stack[nd_index] = data
                # Determine the largest dtype.
                if dtype is None:
                    dtype = data.dtype
                else:
                    dtype = np.promote_types(data.dtype, dtype)

            # Coerce to the largest dtype.
            for nd_index in nd_indexes:
                stack[nd_index] = stack[nd_index].astype(dtype)

            merged_data = multidim_lazy_stack(stack)
            if all_have_data:
                # All inputs were concrete, so turn the result back into a
                # normal array.
                dtype = self._cube_signature.data_type
                merged_data = as_concrete_data(merged_data)
            merged_cube = self._get_cube(merged_data)
            merged_cubes.append(merged_cube)

        return merged_cubes

    def register(self, cube, error_on_mismatch=False):
        """
        Add a compatible :class:`iris.cube.Cube` as a source-cube for
        merging under this :class:`ProtoCube`.

        A cube will be deemed compatible based on the signature of the
        cube and the signature of its scalar coordinates and vector
        coordinates being identical to that of the ProtoCube.

        Args:

        * cube:
            Candidate :class:`iris.cube.Cube` to be associated with
            this :class:`ProtoCube`.

        Kwargs:

        * error_on_mismatch:
            If True, raise an informative
            :class:`~iris.exceptions.MergeError` if registration fails.

        Returns:
            True iff the :class:`iris.cube.Cube` is compatible with
            this :class:`ProtoCube`.

        """
        cube_signature = self._cube_signature
        other = self._build_signature(cube)
        match = cube_signature.match(other, error_on_mismatch)
        if match:
            coord_payload = self._extract_coord_payload(cube)
            match = coord_payload.match_signature(
                self._coord_signature, error_on_mismatch
            )
        if match:
            # Register the cube as a source-cube for this ProtoCube.
            self._add_cube(cube, coord_payload)
        return match

    def _guess_axis(self, name):
        """
        Returns a "best guess" axis name of the candidate dimension.

        Heuristic categoration of the candidate dimension
        (i.e. scalar_defn index) into either label 'T', 'Z', 'Y', 'X'
        or None.

        Based on the associated scalar coordinate definition rather than the
        scalar coordinate itself.

        Args:

        * name:
            The candidate dimension.

        Returns:
            'T', 'Z', 'Y', 'X', or None.

        """
        axis = None

        if not _is_combination(name):
            defn = self._coord_signature.scalar_defns[name]
            axis = iris.util.guess_coord_axis(defn)

        return axis

    def _define_space(self, space, positions, indexes, function_matrix):
        """
        Given the derived :class:`ProtoCube` space, define this space in
        terms of its dimensionality, shape, coordinates and associated
        coordinate to space dimension mappings.

        Args:

        * space:
            A dictionary defining for each candidate dimension its
            dependency on any other candidate dimensions within the space.

        * positions:
            A list containing a dictionary of candidate dimension key to
            scalar value pairs for each source-cube.

        * indexes:
            A cross-reference dictionary for each candidate dimension.

        * function_matrix:
            The function mapping dictionary for each candidate dimension that
            participates in a functional relationship.

        """
        # Heuristic reordering of coordinate defintion indexes into
        # preferred dimension order.
        def axis_and_name(name):
            axis_dict = {"T": 1, "Z": 2, "Y": 3, "X": 4}
            axis_index = axis_dict.get(self._guess_axis(name), 0)
            # The middle element ensures sorting is the same as Python 2.
            return (axis_index, not isinstance(name, int), name)

        names = sorted(space, key=axis_and_name)
        dim_by_name = {}

        metadata = self._coord_metadata
        coord_signature = self._coord_signature
        defns = coord_signature.scalar_defns
        vector_dim_coords_and_dims = coord_signature.vector_dim_coords_and_dims
        vector_aux_coords_and_dims = coord_signature.vector_aux_coords_and_dims
        signature = self._cube_signature

        # First pass - Build the dimension coordinate templates for the space.
        for name in names:
            if space[name] is None:
                if _is_combination(name):
                    members = name.split(_COMBINATION_JOIN)
                    # Create list of unique tuples from all combinations of
                    # scalars for each source cube. The keys of an OrderedDict
                    # are used to retain the ordering of source cubes but to
                    # remove any duplicate tuples.
                    cells = OrderedDict(
                        (
                            tuple(
                                position[
                                    int(member) if member.isdigit() else member
                                ]
                                for member in members
                            ),
                            None,
                        )
                        for position in positions
                    ).keys()
                    dim_by_name[name] = len(self._shape)
                    self._nd_names.append(name)
                    self._shape.append(len(cells))
                    self._stack_shape.append(len(cells))
                    self._cache_by_name[name] = {
                        cell: index for index, cell in enumerate(cells)
                    }
                else:
                    # TODO: Consider appropriate sort order (ascending,
                    # decending) i.e. use CF positive attribute.
                    cells = sorted(indexes[name])
                    points = np.array(
                        [cell.point for cell in cells],
                        dtype=metadata[name].points_dtype,
                    )
                    if cells[0].bound is not None:
                        bounds = np.array(
                            [cell.bound for cell in cells],
                            dtype=metadata[name].bounds_dtype,
                        )
                    else:
                        bounds = None
                    kwargs = dict(zip(CoordMetadata._fields, defns[name]))
                    kwargs.update(metadata[name].kwargs)

                    def name_in_independents():
                        return any(
                            name in independents
                            for independents in space.values()
                            if independents is not None
                        )

                    if len(cells) == 1 and not name_in_independents():
                        # A scalar coordinate not participating in a
                        # function dependency.
                        self._aux_templates.append(
                            _Template((), points, bounds, kwargs)
                        )
                    else:
                        # Dimension coordinate (or aux if the data is
                        # string like).
                        dim_by_name[name] = dim = len(self._shape)
                        self._nd_names.append(name)
                        if metadata[name].points_dtype.kind in "SU":
                            self._aux_templates.append(
                                _Template(dim, points, bounds, kwargs)
                            )
                        else:
                            self._dim_templates.append(
                                _Template(dim, points, bounds, kwargs)
                            )
                        self._shape.append(len(cells))
                        self._stack_shape.append(len(cells))
                        self._cache_by_name[name] = {
                            cell: index for index, cell in enumerate(cells)
                        }

        # Second pass - Build the auxiliary coordinate templates for the space.
        for name in names:
            name_independents = space[name]

            # Determine if there is a function dependency.
            if name_independents is not None:
                # Calculate the auxiliary coordinate shape.
                dims = tuple(
                    [
                        dim_by_name[independent]
                        for independent in name_independents
                    ]
                )
                aux_shape = [self._shape[dim] for dim in dims]
                # Create empty points and bounds in preparation to be filled.
                points = np.empty(aux_shape, dtype=metadata[name].points_dtype)
                if positions[0][name].bound is not None:
                    shape = aux_shape + [len(positions[0][name].bound)]
                    bounds = np.empty(shape, dtype=metadata[name].bounds_dtype)
                else:
                    bounds = None

                # Populate the points and bounds based on the appropriate
                # function mapping.
                temp = function_matrix[name].items()
                for function_independents, name_value in temp:
                    # Build the index (and cache it) for the auxiliary
                    # coordinate based on the associated independent
                    # dimension coordinate/s.
                    index = []

                    name_function_pairs = zip(
                        name_independents, function_independents
                    )
                    for independent, independent_value in name_function_pairs:
                        cache = self._cache_by_name[independent]
                        index.append(cache[independent_value])

                    index = tuple(index)
                    if points is not None:
                        points[index] = name_value.point

                    if bounds is not None:
                        bounds[index] = name_value.bound

                kwargs = dict(zip(CoordMetadata._fields, defns[name]))
                self._aux_templates.append(
                    _Template(dims, points, bounds, kwargs)
                )

        # Calculate the dimension mapping for each vector within the space.
        offset = len(self._shape)
        self._vector_dim_coords_dims = [
            tuple([dim + offset for dim in item.dims])
            for item in vector_dim_coords_and_dims
        ]
        self._vector_aux_coords_dims = [
            tuple([dim + offset for dim in item.dims])
            for item in vector_aux_coords_and_dims
        ]

        # Now factor in the vector payload shape. Note that, for
        # deferred loading, this does NOT change the shape.
        self._shape.extend(signature.data_shape)

    def _get_cube(self, data):
        """
        Return a fully constructed cube for the given data, containing
        all its coordinates and metadata.

        """
        signature = self._cube_signature
        dim_coords_and_dims = [
            (deepcopy(coord), dim) for coord, dim in self._dim_coords_and_dims
        ]
        aux_coords_and_dims = [
            (deepcopy(coord), dims)
            for coord, dims in self._aux_coords_and_dims
        ]
        kwargs = dict(zip(CubeMetadata._fields, signature.defn))

        cms_and_dims = [
            (deepcopy(cm), dims) for cm, dims in self._cell_measures_and_dims
        ]
        avs_and_dims = [
            (deepcopy(av), dims)
            for av, dims in self._ancillary_variables_and_dims
        ]
        cube = iris.cube.Cube(
            data,
            dim_coords_and_dims=dim_coords_and_dims,
            aux_coords_and_dims=aux_coords_and_dims,
            cell_measures_and_dims=cms_and_dims,
            ancillary_variables_and_dims=avs_and_dims,
            **kwargs,
        )

        # Add on any aux coord factories.
        for factory_defn in self._coord_signature.factory_defns:
            args = {}
            for key, defn in factory_defn.dependency_defns:
                coord = cube.coord(defn)
                args[key] = coord
            factory = factory_defn.class_(**args)
            cube.add_aux_factory(factory)

        return cube

    def _nd_index(self, position):
        """
        Returns the n-dimensional index of this source-cube (position),
        within the merged cube.

        """

        index = []

        # Determine the index of the source-cube cell for each dimension.
        for name in self._nd_names:
            if _is_combination(name):
                members = name.split(_COMBINATION_JOIN)
                cell = tuple(
                    position[int(member) if member.isdigit() else member]
                    for member in members
                )
                index.append(self._cache_by_name[name][cell])
            else:
                index.append(self._cache_by_name[name][position[name]])

        return tuple(index)

    def _build_coordinates(self):
        """
        Build the dimension and auxiliary coordinates for the final
        merged cube given that the final dimensionality of the target
        merged cube is known and the associated dimension/s that each
        coordinate maps onto in that merged cube.

        The associated vector coordinate/s of the ProtoCube are also
        created with the correct space dimensionality and dimension/s
        mappings.

        """
        # Containers for the newly created coordinates and associated
        # dimension mappings.
        dim_coords_and_dims = self._dim_coords_and_dims
        aux_coords_and_dims = self._aux_coords_and_dims

        # Build the dimension coordinates.
        for template in self._dim_templates:
            # Sometimes it's not possible to build a dim coordinate e.g.
            # the bounds are not monotonic, so try building the coordinate,
            # and if it fails make the coordinate into an auxiliary coordinate.
            # This will ultimately make an anonymous dimension.
            try:
                coord = iris.coords.DimCoord(
                    template.points, bounds=template.bounds, **template.kwargs
                )
                dim_coords_and_dims.append(_CoordAndDims(coord, template.dims))
            except ValueError:
                self._aux_templates.append(template)

        # There is the potential that there are still anonymous dimensions.
        # Get a list of the dimensions which are not anonymous at this stage.
        covered_dims = [
            dim_coord_and_dim.dims for dim_coord_and_dim in dim_coords_and_dims
        ]

        # Build the auxiliary coordinates.
        for template in self._aux_templates:
            # Attempt to build a DimCoord and add it to the cube. If this
            # fails e.g it's non-monontic or multi-dimensional or non-numeric,
            # then build an AuxCoord.
            try:
                coord = iris.coords.DimCoord(
                    template.points, bounds=template.bounds, **template.kwargs
                )
                if (
                    len(template.dims) == 1
                    and template.dims[0] not in covered_dims
                ):
                    dim_coords_and_dims.append(
                        _CoordAndDims(coord, template.dims)
                    )
                    covered_dims.append(template.dims[0])
                else:
                    aux_coords_and_dims.append(
                        _CoordAndDims(coord, template.dims)
                    )
            except ValueError:
                # kwarg not applicable to AuxCoord.
                template.kwargs.pop("circular", None)
                coord = iris.coords.AuxCoord(
                    template.points, bounds=template.bounds, **template.kwargs
                )
                aux_coords_and_dims.append(_CoordAndDims(coord, template.dims))

        # Mix in the vector coordinates.
        for item, dims in zip(
            self._coord_signature.vector_dim_coords_and_dims,
            self._vector_dim_coords_dims,
        ):
            dim_coords_and_dims.append(_CoordAndDims(item.coord, dims))

        for item, dims in zip(
            self._coord_signature.vector_aux_coords_and_dims,
            self._vector_aux_coords_dims,
        ):
            aux_coords_and_dims.append(_CoordAndDims(item.coord, dims))

    def _build_signature(self, cube):
        """
        Generate the signature that defines this cube.

        Args:

        * cube:
            The source cube to create the cube signature from.

        Returns:
            The cube signature.

        """

        return _CubeSignature(
            cube.metadata,
            cube.shape,
            cube.dtype,
            cube._cell_measures_and_dims,
            cube._ancillary_variables_and_dims,
        )

    def _add_cube(self, cube, coord_payload):
        """Create and add the source-cube skeleton to the ProtoCube."""
        skeleton = _Skeleton(coord_payload.scalar.values, cube.core_data())
        # Attempt to do something sensible with mixed scalar dtypes.
        for i, metadata in enumerate(coord_payload.scalar.metadata):
            if metadata.points_dtype > self._coord_metadata[i].points_dtype:
                self._coord_metadata[i] = metadata
        self._skeletons.append(skeleton)

    def _extract_coord_payload(self, cube):
        """
        Extract all relevant coordinate data and metadata from the cube.

        In particular, for each scalar coordinate determine its definition,
        its cell (point and bound) value and all other scalar coordinate
        metadata that allows us to fully reconstruct that scalar
        coordinate. Note that all scalar data is sorted in order of the
        scalar coordinate definition.

        The coordinate payload of the cube also includes any associated vector
        coordinates that describe that cube, and descriptions of any auxiliary
        coordinate factories.

        """
        scalar_defns = []
        scalar_values = []
        scalar_metadata = []
        vector_dim_coords_and_dims = []
        vector_aux_coords_and_dims = []

        cube_aux_coords = cube.aux_coords
        coords = cube.dim_coords + cube_aux_coords
        cube_aux_coord_ids = {id(coord) for coord in cube_aux_coords}

        # Coordinate hint ordering dictionary - from most preferred to least.
        # Copes with duplicate hint entries, where the most preferred is king.
        hint_dict = {
            name: i
            for i, name in zip(
                range(len(self._hints), 0, -1), self._hints[::-1]
            )
        }
        # Coordinate axis ordering dictionary.
        axis_dict = {"T": 0, "Z": 1, "Y": 2, "X": 3}

        # Coordinate sort function.
        # NB. This makes use of two properties which don't end up in
        # the metadata used by scalar_defns: `coord.points.dtype` and
        # `type(coord)`.
        def key_func(coord):
            points_dtype = coord.dtype
            return (
                not np.issubdtype(points_dtype, np.number),
                not isinstance(coord, iris.coords.DimCoord),
                hint_dict.get(coord.name(), len(hint_dict) + 1),
                axis_dict.get(
                    iris.util.guess_coord_axis(coord), len(axis_dict) + 1
                ),
                coord.metadata,
            )

        # Order the coordinates by hints, axis, and definition.
        for coord in sorted(coords, key=key_func):
            if not cube.coord_dims(coord) and coord.shape == (1,):
                # Extract the scalar coordinate data and metadata.
                scalar_defns.append(coord.metadata)
                # Because we know there's a single Cell in the
                # coordinate, it's quicker to roll our own than use
                # Coord.cell().
                points = coord.points
                bounds = coord.bounds
                points_dtype = points.dtype
                if bounds is not None:
                    bounds_dtype = bounds.dtype
                    bounds = bounds[0]
                else:
                    bounds_dtype = None
                scalar_values.append(iris.coords.Cell(points[0], bounds))
                kwargs = {}
                if isinstance(coord, iris.coords.DimCoord):
                    kwargs["circular"] = coord.circular
                scalar_metadata.append(
                    _CoordMetaData(points_dtype, bounds_dtype, kwargs)
                )
            else:
                # Extract the vector coordinate and metadata.
                if id(coord) in cube_aux_coord_ids:
                    vector_aux_coords_and_dims.append(
                        _CoordAndDims(coord, tuple(cube.coord_dims(coord)))
                    )
                else:
                    vector_dim_coords_and_dims.append(
                        _CoordAndDims(coord, tuple(cube.coord_dims(coord)))
                    )

        factory_defns = []
        for factory in sorted(
            cube.aux_factories, key=lambda factory: factory.metadata
        ):
            dependency_defns = []
            dependencies = factory.dependencies
            for key in sorted(dependencies):
                coord = dependencies[key]
                if coord is not None:
                    dependency_defns.append((key, coord.metadata))
            factory_defn = _FactoryDefn(type(factory), dependency_defns)
            factory_defns.append(factory_defn)

        scalar = _ScalarCoordPayload(
            scalar_defns, scalar_values, scalar_metadata
        )
        vector = _VectorCoordPayload(
            vector_dim_coords_and_dims, vector_aux_coords_and_dims
        )

        return _CoordPayload(scalar, vector, factory_defns)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""A collection of helpers for interpolation."""

from collections import namedtuple
from itertools import product
import operator

from numpy.lib.stride_tricks import as_strided
import numpy as np
import numpy.ma as ma

from iris.analysis._scipy_interpolate import _RegularGridInterpolator
from iris.analysis.cartography import wrap_lons as wrap_circular_points
from iris.coords import DimCoord, AuxCoord
import iris.util


_DEFAULT_DTYPE = np.float16


ExtrapolationMode = namedtuple(
    "ExtrapolationMode",
    ["bounds_error", "fill_value", "mask_fill_value", "force_mask"],
)

EXTRAPOLATION_MODES = {
    "extrapolate": ExtrapolationMode(False, None, None, False),
    "error": ExtrapolationMode(True, 0, 0, False),
    "nan": ExtrapolationMode(False, np.nan, 0, False),
    "mask": ExtrapolationMode(False, np.nan, 1, True),
    "nanmask": ExtrapolationMode(False, np.nan, 1, False),
}


def _canonical_sample_points(coords, sample_points):
    """
    Return the canonical form of the points values.

    Ensures that any points supplied as datetime objects, or similar,
    are converted to their numeric form.

    """
    canonical_sample_points = []
    for coord, points in zip(coords, sample_points):
        if coord.units.is_time_reference():

            def convert_date(date):
                try:
                    date = coord.units.date2num(date)
                except AttributeError:
                    pass
                return date

            convert_dates = np.vectorize(convert_date, [np.dtype(float)])
            points = convert_dates(points)
        canonical_sample_points.append(points)
    return canonical_sample_points


def extend_circular_coord(coord, points):
    """
    Return coordinates points with a shape extended by one
    This is common when dealing with circular coordinates.

    """
    modulus = np.array(coord.units.modulus or 0, dtype=coord.dtype)
    points = np.append(points, points[0] + modulus)
    return points


def extend_circular_coord_and_data(coord, data, coord_dim):
    """
    Return coordinate points and a data array with a shape extended by one
    in the coord_dim axis. This is common when dealing with circular
    coordinates.

    """
    points = extend_circular_coord(coord, coord.points)
    data = extend_circular_data(data, coord_dim)
    return points, data


def extend_circular_data(data, coord_dim):
    coord_slice_in_cube = [slice(None)] * data.ndim
    coord_slice_in_cube[coord_dim] = slice(0, 1)

    mod = ma if ma.isMaskedArray(data) else np
    data = mod.concatenate(
        (data, data[tuple(coord_slice_in_cube)]), axis=coord_dim
    )
    return data


def get_xy_dim_coords(cube):
    """
    Return the x and y dimension coordinates from a cube.

    This function raises a ValueError if the cube does not contain one and
    only one set of x and y dimension coordinates. It also raises a ValueError
    if the identified x and y coordinates do not have coordinate systems that
    are equal.

    Args:

    * cube:
        An instance of :class:`iris.cube.Cube`.

    Returns:
        A tuple containing the cube's x and y dimension coordinates.

    """
    return get_xy_coords(cube, dim_coords=True)


def get_xy_coords(cube, dim_coords=False):
    """
    Return the x and y coordinates from a cube.

    This function raises a ValueError if the cube does not contain one and
    only one set of x and y coordinates. It also raises a ValueError
    if the identified x and y coordinates do not have coordinate systems that
    are equal.

    Args:

    * cube:
        An instance of :class:`iris.cube.Cube`.

    Kwargs:

    * dim_coords:
        Set this to True to only return dimension coordinates. Defaults to
        False.

    Returns:
        A tuple containing the cube's x and y dimension coordinates.

    """
    x_coords = cube.coords(axis="x", dim_coords=dim_coords)
    if len(x_coords) != 1 or x_coords[0].ndim != 1:
        raise ValueError(
            "Cube {!r} must contain a single 1D x "
            "coordinate.".format(cube.name())
        )
    x_coord = x_coords[0]

    y_coords = cube.coords(axis="y", dim_coords=dim_coords)
    if len(y_coords) != 1 or y_coords[0].ndim != 1:
        raise ValueError(
            "Cube {!r} must contain a single 1D y "
            "coordinate.".format(cube.name())
        )
    y_coord = y_coords[0]

    if x_coord.coord_system != y_coord.coord_system:
        raise ValueError(
            "The cube's x ({!r}) and y ({!r}) "
            "coordinates must have the same coordinate "
            "system.".format(x_coord.name(), y_coord.name())
        )

    return x_coord, y_coord


def snapshot_grid(cube):
    """
    Helper function that returns deep copies of lateral (dimension) coordinates
    from a cube.

    """
    x, y = get_xy_dim_coords(cube)
    return x.copy(), y.copy()


class RectilinearInterpolator:
    """
    This class provides support for performing nearest-neighbour or
    linear interpolation over one or more orthogonal dimensions.

    """

    def __init__(self, src_cube, coords, method, extrapolation_mode):
        """
        Perform interpolation over one or more orthogonal coordinates.

        Args:

        * src_cube:
            The :class:`iris.cube.Cube` which is to be interpolated.
        * coords:
            The names or coordinate instances which are to be
            interpolated over
        * method:
            Either 'linear' or 'nearest'.
        * extrapolation_mode:
            Must be one of the following strings:

              * 'extrapolate' - The extrapolation points will be calculated
                according to the method. The 'linear' method extends the
                gradient of the closest two points. The 'nearest' method
                uses the value of the closest point.
              * 'nan' - The extrapolation points will be be set to NaN.
              * 'error' - A ValueError exception will be raised, notifying an
                attempt to extrapolate.
              * 'mask' - The extrapolation points will always be masked, even
                if the source data is not a MaskedArray.
              * 'nanmask' - If the source data is a MaskedArray the
                extrapolation points will be masked. Otherwise they will be
                set to NaN.

        """
        # Trigger any deferred loading of the source cube's data and snapshot
        # its state to ensure that the interpolator is impervious to external
        # changes to the original source cube. The data is loaded to prevent
        # the snaphot having lazy data, avoiding the potential for the
        # same data to be loaded again and again.
        if src_cube.has_lazy_data():
            src_cube.data
        self._src_cube = src_cube.copy()
        # Coordinates defining the dimensions to be interpolated.
        self._src_coords = [self._src_cube.coord(coord) for coord in coords]
        # Whether to use linear or nearest-neighbour interpolation.
        if method not in ("linear", "nearest"):
            msg = "Interpolation method {!r} not supported".format(method)
            raise ValueError(msg)
        self._method = method
        # The extrapolation mode.
        if extrapolation_mode not in EXTRAPOLATION_MODES:
            msg = "Extrapolation mode {!r} not supported."
            raise ValueError(msg.format(extrapolation_mode))
        self._mode = extrapolation_mode
        # The point values defining the dimensions to be interpolated.
        self._src_points = []
        # A list of flags indicating dimensions that need to be reversed.
        self._coord_decreasing = []
        # The cube dimensions to be interpolated over.
        self._interp_dims = []
        # meta-data to support circular data-sets.
        self._circulars = []
        # Instance of the interpolator that performs the actual interpolation.
        self._interpolator = None

        # Perform initial start-up configuration and validation.
        self._setup()

    @property
    def cube(self):
        return self._src_cube

    @property
    def coords(self):
        return self._src_coords

    @property
    def method(self):
        return self._method

    @property
    def extrapolation_mode(self):
        return self._mode

    def _account_for_circular(self, points, data):
        """
        Extend the given data array, and re-centralise coordinate points
        for circular (1D) coordinates.

        """
        for (circular, modulus, index, dim, offset) in self._circulars:
            if modulus:
                # Map all the requested values into the range of the source
                # data (centred over the centre of the source data to allow
                # extrapolation where required).
                points[:, index] = wrap_circular_points(
                    points[:, index], offset, modulus
                )

            # Also extend data if circular (to match the coord points, which
            # 'setup' already extended).
            if circular:
                data = extend_circular_data(data, dim)

        return points, data

    def _account_for_inverted(self, data):
        if np.any(self._coord_decreasing):
            dim_slices = [slice(None)] * data.ndim
            for interp_dim, flip in zip(
                self._interp_dims, self._coord_decreasing
            ):
                if flip:
                    dim_slices[interp_dim] = slice(-1, None, -1)
            data = data[tuple(dim_slices)]
        return data

    def _interpolate(self, data, interp_points):
        """
        Interpolate a data array over N dimensions.

        Create and cache the underlying interpolator instance before invoking
        it to perform interpolation over the data at the given coordinate point
        values.

        * data (ndarray):
            A data array, to be interpolated in its first 'N' dimensions.

        * interp_points (ndarray):
            An array of interpolation coordinate values.
            Its shape is (..., N) where N is the number of interpolation
            dimensions.
            "interp_points[..., i]" are interpolation point values for the i'th
            coordinate, which is mapped to the i'th data dimension.
            The other (leading) dimensions index over the different required
            sample points.

        Returns:

            A :class:`np.ndarray`.  Its shape is "points_shape + extra_shape",
            where "extra_shape" is the remaining non-interpolated dimensions of
            the data array (i.e. 'data.shape[N:]'), and "points_shape" is the
            leading dimensions of interp_points,
            (i.e. 'interp_points.shape[:-1]').

        """
        dtype = self._interpolated_dtype(data.dtype)
        if data.dtype != dtype:
            # Perform dtype promotion.
            data = data.astype(dtype)

        mode = EXTRAPOLATION_MODES[self._mode]
        if self._interpolator is None:
            # Cache the interpolator instance.
            # NB. The constructor of the _RegularGridInterpolator class does
            # some unnecessary checks on the fill_value parameter,
            # so we set it afterwards instead. Sneaky. ;-)
            self._interpolator = _RegularGridInterpolator(
                self._src_points,
                data,
                method=self.method,
                bounds_error=mode.bounds_error,
                fill_value=None,
            )
        else:
            self._interpolator.values = data

        # We may be re-using a cached interpolator, so ensure the fill
        # value is set appropriately for extrapolating data values.
        self._interpolator.fill_value = mode.fill_value
        result = self._interpolator(interp_points)

        if result.dtype != data.dtype:
            # Cast the data dtype to be as expected. Note that, the dtype
            # of the interpolated result is influenced by the dtype of the
            # interpolation points.
            result = result.astype(data.dtype)

        if np.ma.isMaskedArray(data) or mode.force_mask:
            # NB. np.ma.getmaskarray returns an array of `False` if
            # `data` is not a masked array.
            src_mask = np.ma.getmaskarray(data)
            # Switch the extrapolation to work with mask values.
            self._interpolator.fill_value = mode.mask_fill_value
            self._interpolator.values = src_mask
            mask_fraction = self._interpolator(interp_points)
            new_mask = mask_fraction > 0
            if ma.isMaskedArray(data) or np.any(new_mask):
                result = np.ma.MaskedArray(result, new_mask)

        return result

    def _resample_coord(self, sample_points, coord, coord_dims):
        """
        Interpolate the given coordinate at the provided sample points.

        """
        # NB. This section is ripe for improvement:
        # - Internally self._points() expands coord.points to the same
        #   N-dimensional shape as the cube's data, but it doesn't
        #   collapse it again before returning so we have to do that
        #   here.
        # - By expanding to N dimensions self._points() is doing
        #   unnecessary work.
        data = self._points(sample_points, coord.points, coord_dims)
        index = tuple(
            0 if dim not in coord_dims else slice(None)
            for dim in range(self._src_cube.ndim)
        )
        new_points = data[index]
        # Watch out for DimCoord instances that are no longer monotonic
        # after the resampling.
        try:
            new_coord = coord.copy(new_points)
        except (ValueError, TypeError):
            aux_coord = AuxCoord.from_coord(coord)
            new_coord = aux_coord.copy(new_points)
        return new_coord

    def _setup(self):
        """
        Perform initial start-up configuration and validation based on the
        cube and the specified coordinates to be interpolated over.

        """
        # Pre-calculate control data for each interpolation coordinate.
        self._src_points = []
        self._coord_decreasing = []
        self._circulars = []
        self._interp_dims = []
        for index, coord in enumerate(self._src_coords):
            coord_dims = self._src_cube.coord_dims(coord)
            coord_points = coord.points

            # Record if coord is descending-order, and adjust points.
            # (notes copied from pelson :-
            #    Force all coordinates to be monotonically increasing.
            #    Generally this isn't always necessary for a rectilinear
            #    interpolator, but it is a common requirement.)
            decreasing = (
                coord.ndim == 1
                and
                # NOTE: this clause avoids an error when > 1D,
                # as '_validate' raises a more readable error.
                coord_points.size > 1
                and coord_points[1] < coord_points[0]
            )
            self._coord_decreasing.append(decreasing)
            if decreasing:
                coord_points = coord_points[::-1]

            # Record info if coord is circular, and adjust points.
            circular = getattr(coord, "circular", False)
            modulus = getattr(coord.units, "modulus", 0)
            if circular or modulus:
                # Only DimCoords can be circular.
                if circular:
                    coord_points = extend_circular_coord(coord, coord_points)
                offset = 0.5 * (
                    coord_points.max() + coord_points.min() - modulus
                )
                self._circulars.append(
                    (circular, modulus, index, coord_dims[0], offset)
                )

            self._src_points.append(coord_points)

            # Record any interpolation cube dims we haven't already seen.
            coord_dims = [c for c in coord_dims if c not in self._interp_dims]
            self._interp_dims += coord_dims

        self._validate()

    def _validate(self):
        """
        Perform all sanity checks to ensure that the interpolation request
        over the cube with the specified coordinates is valid and can be
        performed.

        """
        if len(set(self._interp_dims)) != len(self._src_coords):
            raise ValueError(
                "Coordinates repeat a data dimension - the "
                "interpolation would be over-specified."
            )

        for coord in self._src_coords:
            if coord.ndim != 1:
                raise ValueError(
                    "Interpolation coords must be 1-d for "
                    "rectilinear interpolation."
                )

            if not isinstance(coord, DimCoord):
                # Check monotonic.
                if not iris.util.monotonic(coord.points, strict=True):
                    msg = (
                        "Cannot interpolate over the non-"
                        "monotonic coordinate {}."
                    )
                    raise ValueError(msg.format(coord.name()))

    def _interpolated_dtype(self, dtype):
        """
        Determine the minimum base dtype required by the
        underlying interpolator.

        """
        if self._method == "nearest":
            result = dtype
        else:
            result = np.result_type(_DEFAULT_DTYPE, dtype)
        return result

    def _points(self, sample_points, data, data_dims=None):
        """
        Interpolate the given data values at the specified list of orthogonal
        (coord, points) pairs.

        Args:

        * sample_points:
            A list of N iterables, where N is the number of coordinates
            passed to the constructor.
            [sample_values_for_coord_0, sample_values_for_coord_1, ...]
        * data:
            The data to interpolate - not necessarily the data from the cube
            that was used to construct this interpolator. If the data has
            fewer dimensions, then data_dims must be defined.

        Kwargs:

        * data_dims:
            The dimensions of the given data array in terms of the original
            cube passed through to this interpolator's constructor. If None,
            the data dimensions must map one-to-one onto the increasing
            dimension order of the cube.

        Returns:
            An :class:`~numpy.ndarray` or :class:`~numpy.ma.MaskedArray`
            instance of the interpolated data.

        """
        dims = list(range(self._src_cube.ndim))
        data_dims = data_dims or dims

        if len(data_dims) != data.ndim:
            msg = (
                "Data being interpolated is not consistent with "
                "the data passed through."
            )
            raise ValueError(msg)

        if sorted(data_dims) != list(data_dims):
            # To do this, a pre & post transpose will be necessary.
            msg = "Currently only increasing data_dims is supported."
            raise NotImplementedError(msg)

        # Broadcast the data into the shape of the original cube.
        if data_dims != list(range(self._src_cube.ndim)):
            strides = list(data.strides)
            for dim in range(self._src_cube.ndim):
                if dim not in data_dims:
                    strides.insert(dim, 0)
            data = as_strided(
                data, strides=strides, shape=self._src_cube.shape
            )

        data = self._account_for_inverted(data)
        # Calculate the transpose order to shuffle the interpolated dimensions
        # to the lower dimensions for the interpolation algorithm. Then the
        # transpose order to restore the dimensions to their original
        # positions.
        di = self._interp_dims
        ds = sorted(dims, key=lambda d: d not in di)
        dmap = {d: di.index(d) if d in di else ds.index(d) for d in dims}
        interp_order, _ = zip(
            *sorted(dmap.items(), key=operator.itemgetter(1))
        )
        _, src_order = zip(*sorted(dmap.items(), key=operator.itemgetter(0)))

        # Prepare the sample points for interpolation and calculate the
        # shape of the interpolated result.
        interp_points = []
        interp_shape = []
        for index, points in enumerate(sample_points):
            dtype = self._interpolated_dtype(self._src_points[index].dtype)
            points = np.array(points, dtype=dtype, ndmin=1)
            interp_points.append(points)
            interp_shape.append(points.size)

        interp_shape.extend(
            length for dim, length in enumerate(data.shape) if dim not in di
        )

        # Convert the interpolation points into a cross-product array
        # with shape (n_cross_points, n_dims)
        interp_points = np.asarray([pts for pts in product(*interp_points)])

        # Adjust for circularity.
        interp_points, data = self._account_for_circular(interp_points, data)

        if interp_order != dims:
            # Transpose data in preparation for interpolation.
            data = np.transpose(data, interp_order)

        # Interpolate and reshape the data ...
        result = self._interpolate(data, interp_points)
        result = result.reshape(interp_shape)

        if src_order != dims:
            # Restore the interpolated result to the original
            # source cube dimensional order.
            result = np.transpose(result, src_order)

        return result

    def __call__(self, sample_points, collapse_scalar=True):
        """
        Construct a cube from the specified orthogonal interpolation points.

        Args:

        * sample_points:
            A list of N iterables, where N is the number of coordinates
            passed to the constructor.
            [sample_values_for_coord_0, sample_values_for_coord_1, ...]

        Kwargs:

        * collapse_scalar:
            Whether to collapse the dimension of the scalar sample points
            in the resulting cube. Default is True.

        Returns:
            A cube interpolated at the given sample points. The dimensionality
            of the cube will be the number of original cube dimensions minus
            the number of scalar coordinates, if collapse_scalar is True.

        """
        if len(sample_points) != len(self._src_coords):
            msg = "Expected sample points for {} coordinates, got {}."
            raise ValueError(
                msg.format(len(self._src_coords), len(sample_points))
            )

        sample_points = _canonical_sample_points(
            self._src_coords, sample_points
        )

        data = self._src_cube.data
        # Interpolate the cube payload.
        interpolated_data = self._points(sample_points, data)

        if collapse_scalar:
            # When collapse_scalar is True, keep track of the dimensions for
            # which sample points is scalar : We will remove these dimensions
            # later on.
            _new_scalar_dims = []
            for dim, points in zip(self._interp_dims, sample_points):
                if np.array(points).ndim == 0:
                    _new_scalar_dims.append(dim)

        cube = self._src_cube
        new_cube = iris.cube.Cube(interpolated_data)
        new_cube.metadata = cube.metadata

        def construct_new_coord_given_points(coord, points):
            # Handle what was previously a DimCoord which may no longer be
            # monotonic.
            try:
                return coord.copy(points)
            except ValueError:
                return AuxCoord.from_coord(coord).copy(points)

        # Keep track of id(coord) -> new_coord for aux factory construction
        # later on.
        coord_mapping = {}
        dims_with_dim_coords = []

        def construct_new_coord(coord):
            dims = cube.coord_dims(coord)
            if coord in self._src_coords:
                index = self._src_coords.index(coord)
                new_points = sample_points[index]
                new_coord = construct_new_coord_given_points(coord, new_points)
                # isinstance not possible here as a dimension coordinate can be
                # mapped to the aux coordinates of a cube.
                if coord in cube.aux_coords:
                    dims = [self._interp_dims[index]]
            else:
                if set(dims).intersection(set(self._interp_dims)):
                    # Interpolate the coordinate payload.
                    new_coord = self._resample_coord(
                        sample_points, coord, dims
                    )
                else:
                    new_coord = coord.copy()
            return new_coord, dims

        def gen_new_cube():
            if (
                isinstance(new_coord, DimCoord)
                and len(dims) > 0
                and dims[0] not in dims_with_dim_coords
            ):
                new_cube._add_unique_dim_coord(new_coord, dims)
                dims_with_dim_coords.append(dims[0])
            else:
                new_cube._add_unique_aux_coord(new_coord, dims)
            coord_mapping[id(coord)] = new_coord

        # Copy/interpolate the coordinates.
        for coord in cube.dim_coords + cube.aux_coords:
            new_coord, dims = construct_new_coord(coord)
            gen_new_cube()

        for factory in self._src_cube.aux_factories:
            new_cube.add_aux_factory(factory.updated(coord_mapping))

        if collapse_scalar and _new_scalar_dims:
            dim_slices = [
                0 if dim in _new_scalar_dims else slice(None)
                for dim in range(new_cube.ndim)
            ]
            new_cube = new_cube[tuple(dim_slices)]

        return new_cube
